## @ingroup Optimization-Package_Setups-TRMM
# Trust_Region_Optimization.py
#
# Created:  Apr 2017, T. MacDonald
# Modified: Jun 2017, T. MacDonald
#           Oct 2019, T. MacDonald

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

import numpy as np
try:
    import RCAIDE.Framework.Optimization.Packages.pyopt.pyopt_setup
except:
    pass
from RCAIDE.Framework.Core import Data
from RCAIDE.Framework.Optimization.Common import helper_functions as help_fun
import os
import sys
from scipy.optimize import minimize

# ----------------------------------------------------------------------
#  Trust Region Optimization Class
# ----------------------------------------------------------------------
## @ingroup Optimization-Package_Setups-TRMM
class Trust_Region_Optimization(Data):
    """A trust region optimization
    
    Assumptions:
    Only SNOPT is implemented
    
    Source:
    None
    """      
        
    def __defaults__(self):
        """This sets the default values.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        None

        Outputs:
        None

        Properties Used:
        None
        """          
        
        self.tag                                = 'TR_Opt'
        self.trust_region_max_iterations        = 30
        self.optimizer_max_iterations           = 30
        self.convergence_tolerance              = 1E-6
        self.optimizer_convergence_tolerance    = 1E-6  # used in SNOPT
        self.optimizer_constraint_tolerance     = 1E-6  # used in SNOPT
        self.difference_interval                = 1E-6 
        self.optimizer_function_precision       = 1E-12 # used in SNOPT
        self.trust_region_function_precision    = 1E-12
        self.optimizer_verify_level             = 0
        self.fidelity_levels                    = 2     # only two are currently supported
        self.evaluation_order                   = [1,2] # currently this order is necessary for proper functionality   
        self.optimizer                          = 'SLSQP'
        
    def optimize(self,problem,print_output=False):
        """Optimizes the problem

        Assumptions:
        Currently only works with SNOPT

        Source:
        "A trust-region framework for managing the use of approximation models in optimization," Alexandrov et. al., 1998
        Details do not follow exactly.

        Inputs:
        problem.                 <Nexus class> (also passed into other functions)
          optimization_problem. 
            inputs               Numpy array matching standard RCAIDE optimization setup
            objective            Numpy array matching standard RCAIDE optimization setup
            constraints          Numpy array matching standard RCAIDE optimization setup
          fidelity_level         [-]
        print_output             <boolean> Determines if output is printed during the optimization run

        Outputs:
        (fOpt_corr,xOpt_corr,str):
            fOpt_corr            <float>
            xOpt_corr            <numpy array>
            str                  Varies depending on the result of the optimization
            

        Properties Used:
        self.
          trust_region_max_iterations         [-]    
          fidelity_levels                     [-]
          evaluation_order                    List of the fidelity level order
          evaluate_model(..)
          calculate_correction(..)
          calculate_constraint_violation(..)
          optimizer                           <string> Determines what optimizer is used
          evaluate_corrected_model(..)
          optimizer_max_iterations            [-]
          optimizer_convergence_tolerance     [-]
          optimizer_constraint_tolerance      [-]
          optimizer_function_precision        [-]
          optimizer_verify_level              Int determining if SNOPT will verify that the minimum is level
          accuracy_ratio(..)
          update_tr_size(..)
          convergance_tolerance               [-]
        """          
        if print_output == False:
            devnull = open(os.devnull,'w')
            sys.stdout = devnull
            
        # History writing
        f_out = open('TRM_hist.txt','w')
        import datetime
        f_out.write(str(datetime.datetime.now())+'\n')       
        
        inp = problem.optimization_problem.inputs
        obj = problem.optimization_problem.objective
        con = problem.optimization_problem.constraints 
        tr  = problem.trust_region
        
        # Set inputs
        nam = inp[:,0] # names
        ini = inp[:,1] # initials
        bnd = inp[:,2] # x bounds
        scl = inp[:,3] # x scale
        typ = inp[:,4] # type
    
        (x,scaled_constraints,x_low_bound,x_up_bound,con_up_edge,con_low_edge,name) = self.scale_vals(inp, con, ini, bnd, scl)
        
        # ---------------------------
        # Trust region specific code
        # ---------------------------
        
        iterations = 0
        max_iterations = self.trust_region_max_iterations
        x = np.array(x,dtype='float')
        tr.center = x
        tr_center = x # trust region center
        x_initial = x*1.      
        
        while iterations < max_iterations:
            iterations += 1
            
            # History writing
            f_out.write('Iteration ----- ' + str(iterations) + '\n')
            f_out.write('x_center: ' + str(x.tolist()) + '\n')
            f_out.write('tr size  : ' + str(tr.size) + '\n')   
            
            f    = [None]*self.fidelity_levels
            df   = [None]*self.fidelity_levels
            g    = [None]*self.fidelity_levels
            dg   = [None]*self.fidelity_levels            
            
            for level in self.evaluation_order:
                problem.fidelity_level = level
                res = self.evaluate_model(problem,x)
                f[level-1]  = res[0]    # objective value
                df[level-1] = res[1]    # objective derivate vector
                g[level-1]  = res[2]    # constraints vector
                dg[level-1] = res[3]    # constraints jacobian
                # History writing
                f_out.write('Level    : ' + str(level) + '\n')
                f_out.write('f        : ' + str(res[0][0]) + '\n')
                f_out.write('df       : ' + str(res[1].tolist()) + '\n')
            # assumes high fidelity is last
            f_center = f[-1][0]
                
            # Calculate correction
            corrections = self.calculate_correction(f,df,g,dg,tr)
            
            # Calculate constraint violation
            g_violation_hi_center = self.calculate_constraint_violation(g[-1],con_low_edge,con_up_edge)
            
            # Subproblem
            tr_size = tr.size
            tr.lower_bound = np.max(np.vstack([x_low_bound,x-tr_size]),axis=0)
            tr.upper_bound = np.min(np.vstack([x_up_bound,x+tr_size]),axis=0)      
            
            # Set to base fidelity level for optimizing the corrected model
            problem.fidelity_level = 1
            
            if self.optimizer == 'SNOPT':
                opt_prob = pyOpt.Optimization('RCAIDE',self.evaluate_corrected_model, corrections=corrections,tr=tr)
                
                for ii in range(len(obj)):
                    opt_prob.addObj('f',f_center) 
                for ii in range(0,len(inp)):
                    vartype = 'c'
                    opt_prob.addVar(nam[ii],vartype,lower=tr.lower_bound[ii],upper=tr.upper_bound[ii],value=x[ii])    
                for ii in range(0,len(con)):
                    if con[ii][1]=='<':
                        opt_prob.addCon(name[ii], type='i', upper=con_up_edge[ii])  
                    elif con[ii][1]=='>':
                        opt_prob.addCon(name[ii], type='i', lower=con_low_edge[ii],upper=np.inf)
                    elif con[ii][1]=='=':
                        opt_prob.addCon(name[ii], type='e', equal=con_up_edge[ii])      
                        
                   
                opt = pyOpt.pySNOPT.SNOPT()       
                
                opt.setOption('Major iterations limit'     , self.optimizer_max_iterations)
                opt.setOption('Major optimality tolerance' , self.optimizer_convergence_tolerance)
                opt.setOption('Major feasibility tolerance', self.optimizer_constraint_tolerance)
                opt.setOption('Function precision'         , self.optimizer_function_precision)
                opt.setOption('Verify level'               , self.optimizer_verify_level)           
                
                outputs = opt(opt_prob, sens_type='FD',problem=problem,corrections=corrections,tr=tr)
                
                # output value of 13 indicates that the optimizer could not find an optimum
                if outputs[2]['value'][0] == 13:
                    feasible_flag = False
                else:
                    feasible_flag = True
                fOpt_corr = outputs[0][0]
                xOpt_corr = outputs[1]
                gOpt_corr = np.zeros([1,len(con)])[0]  
                for ii in range(len(con)):
                    gOpt_corr[ii] = opt_prob._solutions[0]._constraints[ii].value  
                    

            elif self.optimizer == 'SLSQP':
                
                bounds = []
                for lb, ub in zip(tr.lower_bound, tr.upper_bound):
                    bounds.append((lb,ub))
                
                constraints = self.initialize_SLSQP_constraints(con,problem,corrections,tr)
                # need corrections, tr
    
                res = minimize(self.evaluate_corrected_model, x, constraints=constraints, \
                               args=(problem,corrections,tr), bounds=bounds)
                
                fOpt_corr = res['fun']
                xOpt_corr = res['x']
                gOpt_corr = problem.all_constraints(xOpt_corr)
                
                if res['success']:
                    feasible_flag = True
                else:
                    feasible_flag = False
                
            else:
                raise ValueError('Selected optimizer not implemented')
            success_flag = feasible_flag            
        
            f_out.write('fopt = ' + str(fOpt_corr)+'\n')
            f_out.write('xopt = ' + str(xOpt_corr)+'\n')
            f_out.write('gopt = ' + str(gOpt_corr)+'\n')
            
            
            # Constraint minization ------------------------------------------------------------------------
            if feasible_flag == False:
                print('Infeasible within trust region, attempting to minimize constraint')
                
                if self.optimizer == 'SNOPT':
                    opt_prob = pyOpt.Optimization('RCAIDE',self.evaluate_constraints, corrections=corrections,tr=tr,
                                                  lb=con_low_edge,ub=con_up_edge)
                    for ii in range(len(obj)):
                        opt_prob.addObj('constraint violation',0.) 
                    for ii in range(0,len(inp)):
                        vartype = 'c'
                        opt_prob.addVar(nam[ii],vartype,lower=tr.lower_bound[ii],upper=tr.upper_bound[ii],value=x[ii])           
                    opt = pyOpt.pySNOPT.SNOPT()            
                    opt.setOption('Major iterations limit'     , self.optimizer_max_iterations)
                    opt.setOption('Major optimality tolerance' , self.optimizer_convergence_tolerance)
                    opt.setOption('Major feasibility tolerance', self.optimizer_constraint_tolerance)
                    opt.setOption('Function precision'         , self.optimizer_function_precision)
                    opt.setOption('Verify level'               , self.optimizer_verify_level)                 
                   
                    con_outputs = opt(opt_prob, sens_type='FD',problem=problem,corrections=corrections,tr=tr,
                                      lb=con_low_edge,ub=con_up_edge)
                    xOpt_corr = con_outputs[1]
                    new_outputs = self.evaluate_corrected_model(x, problem=problem,corrections=corrections,tr=tr)
        
                    fOpt_corr = new_outputs[0][0]
                    gOpt_corr = np.zeros([1,len(con)])[0]   
                    for ii in range(len(con)):
                        gOpt_corr[ii] = new_outputs[1][ii]
                elif self.optimizer == 'SLSQP':
                    bounds = []
                    for lb, ub in zip(con_low_edge, con_up_edge):
                        bounds.append((lb,ub)) 
                        
                    res = minimize(self.evaluate_constraints, x, \
                                   args=(problem,corrections,tr,con_low_edge,con_up_edge), bounds=bounds, method='slsqp')                    
                        
                    xOpt_corr = res['x']
                    new_outputs = self.evaluate_corrected_model(x, problem=problem,corrections=corrections,tr=tr,
                                                                return_cons=True)
                    
                    fOpt_corr = new_outputs[0][0]
                    gOpt_corr = np.zeros([1,len(con)])[0]   
                    for ii in range(len(con)):
                        gOpt_corr[ii] = new_outputs[1][ii]               
                        
                else:
                    raise ValueError('Selected optimizer not implemented')
                
                # Constraint minization end ------------------------------------------------------------------------
                

            print('fOpt_corr = ', fOpt_corr)
            print('xOpt_corr = ', xOpt_corr)
            print('gOpt_corr = ', gOpt_corr)
            
            # Evaluate high-fidelity at optimum
            problem.fidelity_level = np.max(self.fidelity_levels)
            fOpt_hi, gOpt_hi = self.evaluate_model(problem,xOpt_corr,der_flag=False)
            fOpt_hi = fOpt_hi[0]
        
            g_violation_opt_corr = self.calculate_constraint_violation(gOpt_corr,con_low_edge,con_up_edge)
            g_violation_opt_hi = self.calculate_constraint_violation(gOpt_hi,con_low_edge,con_up_edge)
            
            # Calculate ratio
            rho = self.accuracy_ratio(f_center,fOpt_hi, fOpt_corr, g_violation_hi_center, g_violation_opt_hi, 
                                      g_violation_opt_corr,tr)  
            
            # Acceptance Test
            accepted = 0
            if( fOpt_hi < f_center ):
                print('Trust region update accepted since objective value is lower\n')
                accepted = 1
            elif( g_violation_opt_hi < g_violation_hi_center ):
                print('Trust region update accepted since nonlinear constraint violation is lower\n')
                accepted = 1
            else:
                print('Trust region update rejected (filter)\n')        
            
            # Update Trust Region Size
            print(tr)
            tr_action = self.update_tr_size(rho,tr,accepted)  
                
            # Terminate if trust region too small
            if( tr.size < tr.minimum_size ):
                print('Trust region too small')
                f_out.write('Trust region too small')
                f_out.close()
                if print_output == False:
                    sys.stdout = sys.__stdout__                  
                return (fOpt_corr,xOpt_corr,'Trust region too small')
            
            # Terminate if solution is infeasible, no change is detected, and trust region does not expand
            if( success_flag == False and tr_action < 3 and\
                np.sum(np.isclose(xOpt_corr,x,rtol=1e-15,atol=1e-14)) == len(x) ):
                print('Solution infeasible, no improvement can be made')
                f_out.write('Solution infeasible, no improvement can be made')
                f_out.close()
                if print_output == False:
                    sys.stdout = sys.__stdout__                  
                return (fOpt_corr,xOpt_corr,'Solution infeasible')      
            
            # History writing
            f_out.write('x opt    : ' + str(xOpt_corr.tolist()) + '\n')
            f_out.write('low obj  : ' + str(fOpt_corr)          + '\n')
            f_out.write('hi  obj  : ' + str(fOpt_hi)            + '\n')
            
            # Convergence check
            if (accepted==1 and (np.abs(f_center-fOpt_hi) < self.convergence_tolerance)):
                print('Hard convergence reached')
                f_out.write('Hard convergence reached')
                f_out.close()
                if print_output == False:
                    sys.stdout = sys.__stdout__                  
                return (fOpt_corr,xOpt_corr,'convergence reached')            
            
            # Update trust region center
            if accepted == 1:
                x = xOpt_corr*1.
                tr.center = x*1.             
            
            print('Iteration number: ' + str(iterations))
            print('x value: ' + str(x.tolist()))
            print('Objective value: ' + str(fOpt_hi))
        
        f_out.write('Max iteration limit reached')
        f_out.close()
        print('Max iteration limit reached')
        if print_output == False:
            sys.stdout = sys.__stdout__          
        return (fOpt_corr,xOpt_corr,'Max iteration limit reached')
            
        
    def evaluate_model(self,problem,x,der_flag=True):
        """Evaluates the RCAIDE nexus problem. This is often a mission evaluation.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        problem.                 <Nexus class>
          objective(..)
          all_constraints(..)
          finite difference(..)
        x                        <numpy array>
        der_flag                 <boolean>  Determines if finite differencing is done

        Outputs:
        f  - function value      <float>
        df - derivative of f     <numpy array>
        g  - constraint value    <numpy array> (only returned if der_flag is True)
        dg - jacobian of g       <numpy array> (only returned if der_flag is True)


        Properties Used:
        self.difference_interval [-]
        """              
        f  = problem.objective(x)
        g  = problem.all_constraints(x)
        
        if der_flag == False:
            return f,g
        
        # build derivatives
        fd_step = self.difference_interval
        df, dg  = problem.finite_difference(x,diff_interval=fd_step)
        
        return (f,df,g,dg)


    def evaluate_corrected_model(self,x,problem=None,corrections=None,tr=None,return_cons=False):
        """Evaluates the RCAIDE nexus problem and applies corrections to the results.
        
        Assumptions:
        None

        Source:
        N/A

        Inputs:
        problem.                 <Nexus class>
          objective(..)
          all_constraints(..)
        corrections              <tuple> Contains correction factors
        tr.center                <array>

        Outputs:
        obj                      function objective
        cons                     list of contraint values
        fail                     indicates if the evaluation was successful

        Properties Used:
        None
        """              
        obj   = problem.objective(x)
        const = problem.all_constraints(x).tolist()
        fail  = np.array(np.isnan(obj.tolist()) or np.isnan(np.array(const).any())).astype(int)
        
        A, b = corrections
        x0   = tr.center
        
        obj   = obj + np.dot(A[0,:],(x-x0))+b[0]
        const = const + np.matmul(A[1:,:],(x-x0))+b[1:]
        const = const.tolist()
    
        print('Inputs')
        print(x)
        print('Obj')
        print(obj)
        print('Con')
        print(const)
            
        if self.optimizer == 'SNOPT' or return_cons:
            return obj,const,fail
        elif self.optimizer == 'SLSQP':
            return obj
        else:
            raise NotImplemented
    
    
    def evaluate_constraints(self,x,problem=None,corrections=None,tr=None,lb=None,ub=None):
        """Evaluates the RCAIDE nexus problem provides an objective value based on constraint violation.
        Correction factors are applied to the evaluation results.
        
        Assumptions:
        None

        Source:
        N/A

        Inputs:
        problem.                 <Nexus class>
          objective(..)
          all_constraints(..)
        corrections              <tuple> Contains correction factors
        tr.center                <numpy array>
        lb                       <numpy array> lower bounds on the constraints
        up                       <numpy array> upper bounds on the constraints

        Outputs:
        obj_cons                 objective based on constraint violation
        cons                     list of contraint values
        fail                     indicates if the evaluation was successful

        Properties Used:
        self.calculate_constraint_violation(..)
        """            
        obj      = problem.objective(x) # evaluate the problem
        const    = problem.all_constraints(x).tolist()
        fail     = np.array(np.isnan(obj.tolist()) or np.isnan(np.array(const).any())).astype(int)
        
        A, b = corrections
        x0   = tr.center
        
        const = const + np.matmul(A[1:,:],(x-x0))+b[1:]
        const = const.tolist()
        
        # get the objective that matters here
        obj_cons = self.calculate_constraint_violation(const,lb,ub)
        const    = None
        
        print('Inputs')
        print(x)
        print('Cons violation')
        print(obj_cons)         
        if self.optimizer == 'SNOPT':
            return obj_cons,const,fail  
        else:
            return obj_cons
        
        
    def calculate_constraint_violation(self,gval,lb,ub):
        """Calculates the constraint violation using a 2-norm of the violated constraint values.
        
        Assumptions:
        None

        Source:
        N/A

        Inputs:
        gval                     <numpy array> constraint values
        lb                       <numpy array> lower bounds on the constraints
        up                       <numpy array> upper bounds on the constraints

        Outputs:
        constraint violation     [-]

        Properties Used:
        None
        """                  
        gdiff = []
  
        for i in range(len(gval)):
            if len(lb) > 0:
                if( gval[i] < lb[i] ):
                    gdiff.append(lb[i] - gval[i])
            if len(ub) > 0:    
                if( gval[i] > ub[i] ):
                    gdiff.append(gval[i] - ub[i])
    
        return np.linalg.norm(gdiff) # 2-norm of violation  
    
    def calculate_correction(self,f,df,g,dg,tr):
        """Calculates additive correction factors.
        
        Assumptions:
        None

        Source:
        N/A

        Inputs:
        f  - function value      <float>
        df - derivative of f     <numpy array>
        g  - constraint value    <numpy array> (only returned if der_flag is True)
        dg - jacobian of g       <numpy array> (only returned if der_flag is True)

        Outputs:
        corr                     <tuple> correction factors

        Properties Used:
        None
        """             
        nr = 1 + g[0].size
        nc = df[0].size
            
        A = np.empty((nr,nc))
        b = np.empty(nr)
            
        # objective correction
        A[0,:] = df[1] - df[0]
        b[0] = f[1] - f[0]
            
        # constraint corrections
        A[1:,:] = dg[1] - dg[0]
        b[1:] = g[1] - g[0]
            
        corr = (A,b)
        
        return corr   
    
    
    def scale_vals(self,inp,con,ini,bnd,scl):
        """Scales inputs, constraints, and their bounds.
        
        Assumptions:
        None

        Source:
        N/A

        Inputs:
        (all RCAIDE format specific numpy arrays)
        inp                 Design variables
        con                 Constraint limits
        ini                 Initial values
        bnd                 Variable bounds
        scl                 Scaling factors

        Outputs:
        x                   <numpy array> Scaled design variables
        scaled_constraints  <numpy array>
        x_low_bound         <numpy array>
        x_up_bound          <numpy array>
        con_up_edge         <numpy array>
        con_low_edge        <numpy array>
        name                <list of strings> List of variable names

        Properties Used:
        None
        """    
        
        # Pull out the constraints and scale them
        bnd_constraints = help_fun.scale_const_bnds(con)
        scaled_constraints = help_fun.scale_const_values(con,bnd_constraints)

        x            = ini/scl        
        x_low_bound  = []
        x_up_bound   = []
        edge         = []
        name         = []
        con_up_edge  = []
        con_low_edge = []
        
        for ii in range(0,len(inp)):
            x_low_bound.append(bnd[ii][0]/scl[ii])
            x_up_bound.append(bnd[ii][1]/scl[ii])

        for ii in range(0,len(con)):
            name.append(con[ii][0])
            edge.append(scaled_constraints[ii])
            if con[ii][1]=='<':
                con_up_edge.append(edge[ii])
                con_low_edge.append(-np.inf)
            elif con[ii][1]=='>':
                con_up_edge.append(np.inf)
                con_low_edge.append(edge[ii])
            elif con[ii][1]=='=':
                con_up_edge.append(edge[ii])
                con_low_edge.append(edge[ii])
            
        x_low_bound  = np.array(x_low_bound)
        x_up_bound   = np.array(x_up_bound)
        con_up_edge  = np.array(con_up_edge)         
        con_low_edge = np.array(con_low_edge)        
        
        return (x,scaled_constraints,x_low_bound,x_up_bound,con_up_edge,con_low_edge,name)
    
    
    def accuracy_ratio(self,f_center,f_hi,f_corr,g_viol_center,g_viol_hi,g_viol_corr,tr):
        """Compute the trust region accuracy ratio.
        
        Assumptions:
        None

        Source:
        N/A

        Inputs:
        f_center                   Objective value at the center of the trust region
        f_hi                       High-fidelity objective value at the expected optimum
        f_corr                     Corrected low-fidelity objective value at the expected optimum
        g_viol_center              Constraint violation at the center of the trust region
        g_viol_hi                  High-fidelity constraint violation at the expected optimum
        g_viol_corr                Corrected low-fidelity constraint violation at the expected optimum
        tr.evaluation_function(..)

        Outputs:
        rho                        [-] accuracy ratio

        Properties Used:
        self.trust_region_function_precision [-]
        """          
        # center value does not change since the corrected function already matches
        high_fidelity_center  = tr.evaluate_function(f_center,g_viol_center)
        high_fidelity_optimum = tr.evaluate_function(f_hi,g_viol_hi)
        low_fidelity_center   = tr.evaluate_function(f_center,g_viol_center)
        low_fidelity_optimum  = tr.evaluate_function(f_corr,g_viol_corr)
        if ( np.abs(low_fidelity_center-low_fidelity_optimum) < self.trust_region_function_precision):
            rho = 1.
        else:
            rho = (high_fidelity_center-high_fidelity_optimum)/(low_fidelity_center-low_fidelity_optimum) 
            
        return rho
    
    
    def update_tr_size(self,rho,tr,accepted):
        """Updates the trust region size based on the accuracy ratio and if it has been accepted.
        
        Assumptions:
        None

        Source:
        N/A

        Inputs:
        rho                  [-] accuracy ratio
        tr. 
          size               [-]
          contraction_factor [-]
          contract_threshold [-]
          expand_threshold   [-]
          expansion_factor   [-]

        Outputs:
        tr_action            [-] number indicating the type of action done by the trust region

        Properties Used:
        None
        """     
        
        tr_size_previous = tr.size
        tr_action = 0 # 1: shrink, 2: no change, 3: expand
        if( not accepted ): # shrink trust region
            tr.size = tr.size*tr.contraction_factor
            tr_action = 1
            print('Trust region shrunk from %f to %f\n\n' % (tr_size_previous,tr.size))        
        elif( rho < 0. ): # bad fit, shrink trust region
            tr.size = tr.size*tr.contraction_factor
            tr_action = 1
            print('Trust region shrunk from %f to %f\n\n' % (tr_size_previous,tr.size))
        elif( rho <= tr.contract_threshold ): # okay fit, shrink trust region
            tr.size = tr.size*tr.contraction_factor
            tr_action = 1
            print('Trust region shrunk from %f to %f\n\n' % (tr_size_previous,tr.size))
        elif( rho <= tr.expand_threshold ): # pretty good fit, retain trust region
            tr_action = 2
            print('Trust region size remains the same at %f\n\n' % tr.size)
        elif( rho <= 1.25 ): # excellent fit, expand trust region
            tr.size = tr.size*tr.expansion_factor
            tr_action = 3
            print('Trust region expanded from %f to %f\n\n' % (tr_size_previous,tr.size))
        else: # rho > 1.25, okay-bad fit, but good for us, retain trust region
            tr_action = 2
            print('Trust region size remains the same at %f\n\n' % tr.size)        
            
        return tr_action
    
    def initialize_SLSQP_constraints(self,con,problem,corrections,tr):
        # Initialize variables according to SLSQP requirements
        slsqp_con_list = []
        for i,c in enumerate(con):
            c_dict = {}
            if c[1] == '<' or c[1] == '>':
                c_dict['type'] = 'ineq'
            elif c[1] == '=':
                c_dict['type'] = 'eq'
            else:
                raise ValueError('Constraint specification not recognized.')
            c_dict['fun'] = self.unpack_constraints_slsqp
            if c[1] == '<':
                c_dict['args'] = [i,-1,c[2],problem,corrections,tr]
            else:
                c_dict['args'] = [i,1,c[2],problem,corrections,tr]
            slsqp_con_list.append(c_dict)
        
        return slsqp_con_list
    
    def unpack_constraints_slsqp(self,x,con_ind,sign,edge,problem,corrections,tr):
    
        A, b = corrections
        x0   = tr.center
        
        const = problem.all_constraints(x).tolist()
        
        const = const + np.matmul(A[1:,:],(x-x0))+b[1:]
        const = const.tolist()   
        
        con = (const[con_ind]-edge)*sign
        
        return con


# RCAIDE/Methods/Noise/Common/__init__.py
# 

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
 
from .atmospheric_attenuation                            import atmospheric_attenuation
from .background_noise                                   import background_noise 
from .noise_tone_correction                              import noise_tone_correction  
from .decibel_arithmetic                                 import pressure_ratio_to_SPL_arithmetic
from .decibel_arithmetic                                 import SPL_arithmetic
from .convert_to_third_octave_band                       import convert_to_third_octave_band 
from .compute_noise_source_coordinates                   import compute_rotor_point_source_coordinates
from .generate_zero_elevation_microphone_locations       import generate_zero_elevation_microphone_locations
from .generate_terrain_microphone_locations              import generate_terrain_microphone_locations
from .generate_hemisphere_microphone_locations           import generate_hemisphere_microphone_locations
from .compute_relative_noise_evaluation_locations        import compute_relative_noise_evaluation_locations 


# RCAIDE/Library/Methods/Weights/Correlation_Buildups/General_Aviation/compute_payload_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE 
from RCAIDE.Framework.Core import  Units ,  Data 

# ----------------------------------------------------------------------------------------------------------------------
# Payload Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_payload_weight(TOW, empty, num_pax, W_cargo, W_passenger = 225.*Units.lbs,W_baggage = 0.):
    """ 
        Calculate the weight of the payload and the resulting fuel mass
 
        Inputs:
            TOW -                                                    [kilograms]
            W_empty - Operating empty weight of the aircraft        [kilograms]
            num_pax - number of passengers on the aircraft           [dimensionless]
            W_cargo - weight of cargo being carried on the aircraft [kilogram]
            W_passenger - weight of each passenger on the aircraft  [kilograms]
            W_baggage - weight of the baggage for each passenger    [kilograms]
            
            
        Outputs:
            output - a data dictionary with fields:
                payload - weight of the passengers plus baggage and paid cargo [kilograms]
                pax - weight of all the passengers                             [kilograms]
                bag - weight of all the baggage                                [kilograms]
                fuel - weight of the fuel carried                              [kilograms]
                empty - operating empty weight of the aircraft                 [kilograms]

    """     

    # process
    W_pax     = W_passenger * num_pax 
    W_bag     = W_baggage * num_pax
    W_payload = W_pax + W_bag + W_cargo

    # packup outputs
    output = Data()
    output.total        = W_payload
    output.passengers   = W_pax
    output.baggage      = W_bag
    output.cargo        = W_cargo

    return output


# RCAIDE/Methods/Energy/Propulsors/Turbojet_Propulsor/design_turbojet.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  

# RCAIDE Imports     
import RCAIDE 
from RCAIDE.Framework.Mission.Common                                 import Conditions
from RCAIDE.Library.Methods.Propulsors.Converters.Ram                import compute_ram_performance
from RCAIDE.Library.Methods.Propulsors.Converters.Combustor          import compute_combustor_performance
from RCAIDE.Library.Methods.Propulsors.Converters.Compressor         import compute_compressor_performance
from RCAIDE.Library.Methods.Propulsors.Converters.Turbine            import compute_turbine_performance
from RCAIDE.Library.Methods.Propulsors.Converters.Supersonic_Nozzle  import compute_supersonic_nozzle_performance
from RCAIDE.Library.Methods.Propulsors.Converters.Compression_Nozzle import compute_compression_nozzle_performance
from RCAIDE.Library.Methods.Propulsors.Turbojet_Propulsor            import size_core 
from RCAIDE.Library.Methods.Propulsors.Common                        import compute_static_sea_level_performance

# Python package imports   
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------  
#  Design Turbojet
# ----------------------------------------------------------------------------------------------------------------------   
def design_turbojet(turbojet):  
    #check if mach number and temperature are passed
    if(turbojet.design_mach_number==None or turbojet.design_altitude==None):
        
        #raise an error
        raise NameError('The sizing conditions require an altitude and a Mach number')
    
    else:
        #call the atmospheric model to get the conditions at the specified altitude
        atmosphere = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
        atmo_data  = atmosphere.compute_values(turbojet.design_altitude,turbojet.design_isa_deviation)
        planet     = RCAIDE.Library.Attributes.Planets.Earth()
        
        p   = atmo_data.pressure          
        T   = atmo_data.temperature       
        rho = atmo_data.density          
        a   = atmo_data.speed_of_sound    
        mu  = atmo_data.dynamic_viscosity   
    
        # setup conditions
        conditions = RCAIDE.Framework.Mission.Common.Results()
    
        # freestream conditions    
        conditions.freestream.altitude                    = np.atleast_1d(turbojet.design_altitude)
        conditions.freestream.mach_number                 = np.atleast_1d(turbojet.design_mach_number)
        conditions.freestream.pressure                    = np.atleast_1d(p)
        conditions.freestream.temperature                 = np.atleast_1d(T)
        conditions.freestream.density                     = np.atleast_1d(rho)
        conditions.freestream.dynamic_viscosity           = np.atleast_1d(mu)
        conditions.freestream.gravity                     = np.atleast_1d(planet.compute_gravity(turbojet.design_altitude))
        conditions.freestream.isentropic_expansion_factor = np.atleast_1d(turbojet.working_fluid.compute_gamma(T,p))
        conditions.freestream.Cp                          = np.atleast_1d(turbojet.working_fluid.compute_cp(T,p))
        conditions.freestream.R                           = np.atleast_1d(turbojet.working_fluid.gas_specific_constant)
        conditions.freestream.speed_of_sound              = np.atleast_1d(a)
        conditions.freestream.velocity                    = np.atleast_1d(a*turbojet.design_mach_number)
  
    fuel_line                                      = RCAIDE.Library.Components.Energy.Distributors.Fuel_Line()
    segment                                        = RCAIDE.Framework.Mission.Segments.Segment()  
    segment.state.conditions                       = conditions
    segment.state.conditions.energy[fuel_line.tag] = Conditions()
    segment.state.conditions.noise[fuel_line.tag]  = Conditions()
    turbojet.append_operating_conditions(segment) 
    for tag, item in  turbojet.items(): 
        if issubclass(type(item), RCAIDE.Library.Components.Component):
            item.append_operating_conditions(segment,turbojet)
    
    ram                       = turbojet.ram
    inlet_nozzle              = turbojet.inlet_nozzle
    low_pressure_compressor   = turbojet.low_pressure_compressor
    high_pressure_compressor  = turbojet.high_pressure_compressor
    combustor                 = turbojet.combustor
    high_pressure_turbine     = turbojet.high_pressure_turbine
    low_pressure_turbine      = turbojet.low_pressure_turbine
    core_nozzle               = turbojet.core_nozzle

    # unpack component conditions
    turbojet_conditions     = conditions.energy[turbojet.tag]
    ram_conditions          = turbojet_conditions[ram.tag]     
    inlet_nozzle_conditions = turbojet_conditions[inlet_nozzle.tag]
    core_nozzle_conditions  = turbojet_conditions[core_nozzle.tag] 
    lpc_conditions          = turbojet_conditions[low_pressure_compressor.tag]
    hpc_conditions          = turbojet_conditions[high_pressure_compressor.tag]
    lpt_conditions          = turbojet_conditions[low_pressure_turbine.tag]
    hpt_conditions          = turbojet_conditions[high_pressure_turbine.tag]
    combustor_conditions    = turbojet_conditions[combustor.tag] 
     
    # Step 1: Set the working fluid to determine the fluid properties
    ram.working_fluid                             = turbojet.working_fluid

    # Step 2: Compute flow through the ram , this computes the necessary flow quantities and stores it into conditions
    compute_ram_performance(ram,ram_conditions,conditions)

    # Step 3: link inlet nozzle to ram 
    inlet_nozzle_conditions.inputs.stagnation_temperature             = ram_conditions.outputs.stagnation_temperature
    inlet_nozzle_conditions.inputs.stagnation_pressure                = ram_conditions.outputs.stagnation_pressure
    inlet_nozzle_conditions.inputs.static_temperature                 = ram_conditions.outputs.static_temperature
    inlet_nozzle_conditions.inputs.static_pressure                    = ram_conditions.outputs.static_pressure
    inlet_nozzle_conditions.inputs.mach_number                        = ram_conditions.outputs.mach_number
    inlet_nozzle.working_fluid                                        = ram.working_fluid

    # Step 4: Compute flow through the inlet nozzle
    compute_compression_nozzle_performance(inlet_nozzle,inlet_nozzle_conditions,conditions)      

    # Step 5: Link low pressure compressor to the inlet nozzle
    lpc_conditions.inputs.stagnation_temperature  = inlet_nozzle_conditions.outputs.stagnation_temperature
    lpc_conditions.inputs.stagnation_pressure     = inlet_nozzle_conditions.outputs.stagnation_pressure
    lpc_conditions.inputs.static_temperature      = inlet_nozzle_conditions.outputs.static_temperature
    lpc_conditions.inputs.static_pressure         = inlet_nozzle_conditions.outputs.static_pressure
    lpc_conditions.inputs.mach_number             = inlet_nozzle_conditions.outputs.mach_number  
    low_pressure_compressor.working_fluid         = inlet_nozzle.working_fluid

    # Step 6: Compute flow through the low pressure compressor
    compute_compressor_performance(low_pressure_compressor,lpc_conditions,conditions)

    # Step 7: Link the high pressure compressor to the low pressure compressor
    hpc_conditions.inputs.stagnation_temperature = lpc_conditions.outputs.stagnation_temperature
    hpc_conditions.inputs.stagnation_pressure    = lpc_conditions.outputs.stagnation_pressure
    hpc_conditions.inputs.static_temperature     = lpc_conditions.outputs.static_temperature
    hpc_conditions.inputs.static_pressure        = lpc_conditions.outputs.static_pressure
    hpc_conditions.inputs.mach_number            = lpc_conditions.outputs.mach_number  
    high_pressure_compressor.working_fluid       = low_pressure_compressor.working_fluid   

    # Step 8: Compute flow through the high pressure compressor
    compute_compressor_performance(high_pressure_compressor,hpc_conditions,conditions)
   
    # Step 9: Link the combustor to the high pressure compressor
    combustor_conditions.inputs.stagnation_temperature                = hpc_conditions.outputs.stagnation_temperature
    combustor_conditions.inputs.stagnation_pressure                   = hpc_conditions.outputs.stagnation_pressure
    combustor_conditions.inputs.static_temperature                    = hpc_conditions.outputs.static_temperature
    combustor_conditions.inputs.static_pressure                       = hpc_conditions.outputs.static_pressure
    combustor_conditions.inputs.mach_number                           = hpc_conditions.outputs.mach_number  
    combustor.working_fluid                                           = high_pressure_compressor.working_fluid  
    
    # Step 10: Compute flow through the high pressor compressor 
    compute_combustor_performance(combustor,combustor_conditions,conditions)

    hpt_conditions.inputs.stagnation_temperature    = combustor_conditions.outputs.stagnation_temperature
    hpt_conditions.inputs.stagnation_pressure       = combustor_conditions.outputs.stagnation_pressure
    hpt_conditions.inputs.fuel_to_air_ratio         = combustor_conditions.outputs.fuel_to_air_ratio 
    hpt_conditions.inputs.static_temperature        = combustor_conditions.outputs.static_temperature
    hpt_conditions.inputs.static_pressure           = combustor_conditions.outputs.static_pressure
    hpt_conditions.inputs.mach_number               = combustor_conditions.outputs.mach_number       
    hpt_conditions.inputs.compressor                = hpc_conditions.outputs   
    hpt_conditions.inputs.bypass_ratio              = 0.0
    high_pressure_turbine.working_fluid             = combustor.working_fluid 

    # Step 11: Compute flow through the high pressure turbine
    compute_turbine_performance(high_pressure_turbine,hpt_conditions,conditions)
            
    # Step 12: Link the low pressure turbine to the high pressure turbine
    lpt_conditions.inputs.stagnation_temperature     = hpt_conditions.outputs.stagnation_temperature
    lpt_conditions.inputs.stagnation_pressure        = hpt_conditions.outputs.stagnation_pressure 
    lpt_conditions.inputs.static_temperature         = hpt_conditions.outputs.static_temperature
    lpt_conditions.inputs.static_pressure            = hpt_conditions.outputs.static_pressure 
    lpt_conditions.inputs.mach_number                = hpt_conditions.outputs.mach_number  
    low_pressure_turbine.working_fluid               = high_pressure_turbine.working_fluid    
    lpt_conditions.inputs.compressor                 = lpc_conditions.outputs 
    lpt_conditions.inputs.fuel_to_air_ratio          = combustor_conditions.outputs.fuel_to_air_ratio 
    lpt_conditions.inputs.bypass_ratio               = 0.0 

    # Step 16: Compute flow through the low pressure turbine
    compute_turbine_performance(low_pressure_turbine,lpt_conditions,conditions)

    # Step 17: Link the core nozzle to the low pressure turbine
    core_nozzle_conditions.inputs.stagnation_temperature     = lpt_conditions.outputs.stagnation_temperature
    core_nozzle_conditions.inputs.stagnation_pressure        = lpt_conditions.outputs.stagnation_pressure
    core_nozzle_conditions.inputs.static_temperature         = lpt_conditions.outputs.static_temperature
    core_nozzle_conditions.inputs.static_pressure            = lpt_conditions.outputs.static_pressure  
    core_nozzle_conditions.inputs.mach_number                = lpt_conditions.outputs.mach_number   
    core_nozzle.working_fluid                                = low_pressure_turbine.working_fluid 

    # Step 18: Compute flow through the core nozzle
    compute_supersonic_nozzle_performance(core_nozzle,core_nozzle_conditions,conditions)
 
    # Step 19: link the thrust component to the core nozzle
    turbojet_conditions.core_nozzle_area_ratio                   = core_nozzle_conditions.outputs.area_ratio 
    turbojet_conditions.core_nozzle_static_pressure              = core_nozzle_conditions.outputs.static_pressure
    turbojet_conditions.core_nozzle_exit_velocity                = core_nozzle_conditions.outputs.velocity 
    turbojet_conditions.fuel_to_air_ratio                        = combustor_conditions.outputs.fuel_to_air_ratio 
    turbojet_conditions.stag_temp_lpt_exit                       = lpc_conditions.outputs.stagnation_temperature
    turbojet_conditions.stag_press_lpt_exit                      = lpc_conditions.outputs.stagnation_pressure 
    turbojet_conditions.total_temperature_reference              = lpc_conditions.outputs.stagnation_temperature
    turbojet_conditions.total_pressure_reference                 = lpc_conditions.outputs.stagnation_pressure 
    turbojet_conditions.flow_through_core                        = 1.0 #scaled constant to turn on core thrust computation
    turbojet_conditions.flow_through_fan                         = 0.0 #scaled constant to turn on fan thrust computation      
    
    # Step 20: Size the core of the turbojet  
    size_core(turbojet,turbojet_conditions,conditions)
    mass_flow                     = turbojet.mass_flow_rate_design
    turbojet.design_core_massflow = mass_flow  
    
    # Step 21: Static Sea Level Thrust 
    compute_static_sea_level_performance(turbojet)
     
    return      
  


# RCAIDE/Library/Plots/Performance/Aerodynamics/__init__.py
# 

"""RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
from .plot_aircraft_aerodynamics                import plot_aircraft_aerodynamics
from .plot_airfoil_boundary_layer_properties    import plot_airfoil_boundary_layer_properties
from .plot_airfoil_polar_files                  import plot_airfoil_polar_files
from .plot_airfoil_polars                       import plot_airfoil_polars
from .plot_airfoil_surface_forces               import plot_airfoil_surface_forces
from .plot_aerodynamic_coefficients             import plot_aerodynamic_coefficients
from .plot_aerodynamic_forces                   import plot_aerodynamic_forces
from .plot_drag_components                      import plot_drag_components
from .plot_lift_distribution                    import plot_lift_distribution
from .plot_rotor_disc_performance               import plot_rotor_disc_performance
from .plot_rotor_performance                    import plot_rotor_performance    
from .plot_disc_and_power_loading               import plot_disc_and_power_loading
from .plot_rotor_conditions                     import plot_rotor_conditions
from .plot_surface_pressures                    import plot_surface_pressures


# RCAIDE/Framework/Analyses/Analysis.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
from RCAIDE.Framework.Core import Data 
from RCAIDE.Framework.Core import Container as ContainerBase 

# ----------------------------------------------------------------------------------------------------------------------
# Analysis
# ----------------------------------------------------------------------------------------------------------------------  
class Analysis(Data):
    """ RCAIDE.Framework.Analyses.Analysis()
    
        The Top Level Analysis Class
        
            Assumptions:
            None
            
            Source:
            N/A
    """
    def __defaults__(self):
        """This sets the default values and methods for the analysis.
        
                Assumptions:
                None
        
                Source:
                N/A
        
                Inputs:
                None
        
                Outputs:
                None
        
                Properties Used:
                N/A
            """           
        self.tag    = 'analysis'
        self.features = Data()
        self.settings = Data() 
        
    def initialize(self,*args,**kwarg):
        """This is used to initialize the analysis' specific algorithms.
                
                Assumptions:
                None
                
                Source:
                N/A
                
                Inputs:
                None
                
                Outputs:
                None
                
                Properties Used:
                N/A
            """        
        return
    
    def evaluate(self,*args,**kwarg):
        """This is used to execute the analysis' specific algorithms.
                
                Assumptions:
                None
                
                Source:
                N/A
                
                Inputs:
                None
                
                Outputs:
                None
                
                Properties Used:
                N/A
            """             
        raise NotImplementedError
        return Data()
    
    def post_process(self,*args,**kwarg):
        """This is used to post_process 
                
                Assumptions:
                None
                
                Source:
                N/A
                
                Inputs:
                None
                
                Outputs:
                None
                
                Properties Used:
                N/A
            """                
        return 
    
    def __call__(self,*args,**kwarg):
        
        """This is used to set the class' call behavior to the evaluate function.
                        
                Assumptions:
                None
                        
                Source:
                N/A
                        
                Inputs:
                None
                        
                Outputs:
                None
                        
                Properties Used:
                N/A
            """                        
        
        return self.evaluate(*args,**kwarg)
    
    
# ----------------------------------------------------------------------------------------------------------------------  
#  CONFIG CONTAINER
# ----------------------------------------------------------------------------------------------------------------------   
class Container(ContainerBase):
    """ RCAIDE.Framework.Analyses.Analysis.Container()
    
        The Analysis Container Class
        
            Assumptions:
            None
            
            Source:
            N/A
    """ 
        
    def initialize(self,*args,**kwarg):
        """This is used to execute the initialize functions of the analyses
            stored in the container.
                                        
                Assumptions:
                None
                                        
                Source:
                N/A
                                        
                Inputs:
                None
                                        
                Outputs:
                None
                                        
                Properties Used:
                N/A
            """                    
        for tag,analysis in self.items:
            if hasattr(analysis,'initialize'):
                analysis.initialize(*args,**kwarg) 
    
    def evaluate(self,*args,**kwarg):
        """This is used to execute the evaluate functions of the analyses
            stored in the container.
                                                
                Assumptions:
                None
                                                
                Source:
                N/A
                                                
                Inputs:
                None
                                                
                Outputs:
                Results of the Evaluate Functions
                                                
                Properties Used:
                N/A
            """
        results = Data()
        for tag,analysis in self.items(): 
            if hasattr(analysis,'evaluate'):
                result = analysis.evaluate(*args,**kwarg)
            else:
                result = analysis(*args,**kwarg)
            results[tag] = result
        return results
    
    def post_process(self,*args,**kwarg):
        """This is used to execute post processing functions 
                                                
                Assumptions:
                None
                                                
                Source:
                N/A
                                                
                Inputs:
                None
                                                
                Outputs:
                None
                                                
                Properties Used:
                N/A
            """        
        
        for tag,analysis in self.items():
            if hasattr(analysis,'post_process'):
                analysis.post_process(*args,**kwarg)
                
    
    def __call__(self,*args,**kwarg): 
        """This is used to set the class' call behavior to the evaluate functions.
                                                        
                Assumptions:
                None
                                                        
                Source:
                N/A
                                                        
                Inputs:
                None
                                                        
                Outputs:
                None
                                                        
                Properties Used:
                N/A
            """                
        
        return self.evaluate(*args,**kwarg)

    
# ----------------------------------------------------------------------------------------------------------------------   
#  LINKING
# ----------------------------------------------------------------------------------------------------------------------   
Analysis.Container = Container


# RCAIDE/Library/Plots/Geometry/plot_airfoil.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
# RCAIDE imports 
from RCAIDE.Library.Plots.Common import plot_style
from RCAIDE.Library.Methods.Geometry.Airfoil  import import_airfoil_geometry  

# package imports 
import matplotlib.pyplot as plt 

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------  
def plot_airfoil(airfoil_paths,
                 save_figure = False, 
                 save_filename = "Airfoil_Geometry",
                 file_type = ".png", 
                 width = 11, height = 7):
    """This plots all airfoil defined in the list "airfoil_names" 

    Assumptions:
    None

    Source:
    None

    Inputs:
    airfoil_geometry_files   <list of strings>

    Outputs: 
    Plots

    Properties Used:
    N/A	
    """
    # get airfoil coordinate geometry     
    airfoil_geometry = import_airfoil_geometry(airfoil_paths)

    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters)    

    fig  = plt.figure(save_filename)
    fig.set_size_inches(width,height) 
    axis = fig.add_subplot(1,1,1)     
    axis.plot(airfoil_geometry.x_coordinates,airfoil_geometry.y_coordinates, color = ps.color, marker = ps.markers[0], linewidth = ps.line_width) 
    axis.set_xlabel('x')
    axis.set_ylabel('y')    
     
    if save_figure:
        fig.savefig(save_filename.replace("_", " ") + file_type)  
     
    return fig



# RCAIDE/Library/Plots/Energy/plot_battery_module_C_rates.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  

from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots.Common import set_axes, plot_style 
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np  

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------   
def plot_battery_module_C_rates(results,
                        save_figure   = False,
                        show_legend   = True,
                        save_filename = "Battery_Module_C_Rates",
                        file_type     =".png",
                        width         = 8,
                        height        = 6):
    """Plots the module-level conditions of the battery throughout flight.

    Assumptions:
    None

    Source:
    None

    Inputs:
    results.segments.conditions.
        freestream.altitude
        weights.total_mass
        weights.vehicle_mass_rate
        frames.body.thrust_force_vector

    Outputs: 
    Plots

    Properties Used:
    N/A	
    """ 
     
    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters)
     

    fig = plt.figure(save_filename)
    fig.set_size_inches(width,height)
    # get line colors for plots 
    line_colors   = cm.inferno(np.linspace(0,0.9,len(results.segments)))      
    axis_1 = plt.subplot(1,2,1)
    axis_2 = plt.subplot(1,2,2)  
     
    for network in results.segments[0].analyses.energy.vehicle.networks: 
        busses  = network.busses 
        for bus in busses:
            for b_i, battery in enumerate(bus.battery_modules):
                if b_i == 0 or bus.identical_battery_modules == False:
                    for i in range(len(results.segments)):  
                        time                  = results.segments[i].conditions.frames.inertial.time[:,0] / Units.min    
                        battery_conditions    = results.segments[i].conditions.energy[bus.tag].battery_modules[battery.tag]     
                        module_energy         = battery_conditions.energy[:,0]
                        module_volts          = battery_conditions.voltage_under_load[:,0] 
                        module_current        = battery_conditions.current[:,0]  
                        module_battery_amp_hr = (module_energy/ Units.Wh )/module_volts
                        module_C_instant      = module_current/module_battery_amp_hr
                        module_C_nominal      = module_current/np.max(module_battery_amp_hr)   
    
                        if i == 0 and i ==0:              
                            axis_1.plot(time, module_C_instant, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width, label = battery.tag)
                        else: 
                            axis_1.plot(time, module_C_instant, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width)                                
                        axis_1.set_ylabel(r'Inst. C-Rate (C)')
                        axis_1.set_xlabel('Time (mins)')
                        set_axes(axis_1)     
                         
                        axis_2.plot(time, module_C_nominal, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width)
                        axis_2.set_ylabel(r'Nom. C-Rate (C)')
                        axis_2.set_xlabel('Time (mins)')
                        set_axes(axis_2)    
    if show_legend:      
        leg =  fig.legend(bbox_to_anchor=(0.5, 0.95), loc='upper center', ncol = 4)  
    
    # Adjusting the sub-plots for legend 
    fig.tight_layout()
    fig.subplots_adjust(top=0.8) 
    
    # set title of plot 
    title_text   = 'Battery Cell Conditions'       
    fig.suptitle(title_text) 
    
    if save_figure:
        plt.savefig(save_filename + battery.tag + file_type)    
    return fig


    return  
 
                 


# RCAIDE/Framework/Analyses/Noise/Frequency_Domain_Buildup.py
# 
# 
# Created:  Oct 2024, A. Molloy
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
# noise imports
from RCAIDE.Framework.Core import  Data
from RCAIDE.Library.Methods.Noise.Common.background_noise     import background_noise
from RCAIDE.Library.Methods.Noise.Metrics import * 
from RCAIDE.Library.Methods.Noise.Common.generate_zero_elevation_microphone_locations import generate_zero_elevation_microphone_locations 
from RCAIDE.Library.Methods.Noise.Common.generate_terrain_microphone_locations        import generate_terrain_microphone_locations     
from RCAIDE.Library.Methods.Noise.Common.compute_relative_noise_evaluation_locations  import compute_relative_noise_evaluation_locations
from RCAIDE.Library.Methods.Geodesics.compute_point_to_point_geospacial_data          import compute_point_to_point_geospacial_data

# package imports
import numpy as np
from scipy.interpolate                                           import RegularGridInterpolator


# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ---------------------------------------------------------------------------------------------------------------------- 
def post_process_noise_data(results,
                            flight_times = np.array(['06:00:00','06:30:00','07:00:00','07:30:00','08:00:00','08:30:00',
                                                     '09:00:00','09:30:00','10:00:00','10:30:00','11:00:00','11:30:00',
                                                     '12:00:00','12:30:00','13:00:00','13:30:00','14:00:00','14:30:00',
                                                     '15:00:00']),
                            time_period             = ['06:00:00','20:00:00'], 
                            evalaute_noise_metrics  = True): 
    """This translates all noise data into metadata for plotting 
    
    Assumptions:
    None
    
    Source: 
 
    Inputs: results 
         
    Outputs: noise_data
    
    Properties Used:
    N/A
    """

    # Step 1: Unpack settings      
    settings   = results.segments[0].analyses.noise.settings
    n          = settings.number_of_microphone_in_stencil
    N_gm_x     = settings.microphone_x_resolution
    N_gm_y     = settings.microphone_y_resolution    
    noise_data = Data()   
     
    # Step 2: Determing microhpone points where noise is to be computed
    microphone_coordinates =  None
    if settings.topography_file !=  None:
        compute_point_to_point_geospacial_data(settings)
        microphone_locations ,microphone_coordinates = generate_terrain_microphone_locations(settings) 
        noise_data.topography_file                    = settings.topography_file  
        noise_data.microphone_coordinates             = microphone_coordinates.reshape(N_gm_x,N_gm_y,3)     
        noise_data.aircraft_origin_coordinates        = settings.aircraft_origin_coordinates          
        noise_data.aircraft_destination_coordinates   = settings.aircraft_destination_coordinates         
    else:    
        microphone_locations =  generate_zero_elevation_microphone_locations(settings)   
    noise_data.microphone_y_resolution       = N_gm_y
    noise_data.microphone_x_resolution       = N_gm_x              
    noise_data.microphone_locations          = microphone_locations.reshape(N_gm_x,N_gm_y,3)         
    
    # Step 3: Create empty arrays to store noise data 
    N_segs = len(results.segments)
    num_gm_mic      = len(microphone_locations)
    num_noise_time  = settings.noise_times_steps 
    
    # Step 4: Initalize Arrays 
    N_ctrl_pts            = ( N_segs-1) * (num_noise_time -1) + num_noise_time # ensures that noise is computed continuously across segments 
    SPL_dBA               = np.ones((N_ctrl_pts,N_gm_x,N_gm_y))*background_noise()  
    Aircraft_pos          = np.empty((0,3))
    Time                  = np.empty((0))
    mic_locs              = np.zeros((N_ctrl_pts,n))   
 
    idx =  0
    
    # Step 5: loop through segments and store noise 
    for seg in range(N_segs):  
        segment    = results.segments[seg]
        settings   = segment.analyses.noise.settings  
        phi        = settings.noise_hemisphere_phi_angles
        theta      = settings.noise_hemisphere_theta_angles
        conditions = segment.state.conditions  
        time       = conditions.frames.inertial.time[:,0]
        
        # Step 5.1 : Compute relative microhpone locations 
        noise_time,noise_pos,RML,PHI,THETA,num_gm_mic  = compute_relative_noise_evaluation_locations(settings, microphone_locations,segment) 
         
        # Step 5.2: Compute aircraft position and npose at interpolated hemisphere locations
        cpt   = 0 
        if seg == (N_segs - 1):
            noise_time_ = noise_time 
        else:
            noise_time_ = noise_time[:-1]
             
        Aircraft_pos = np.vstack((Aircraft_pos,noise_pos))
        Time         = np.hstack((Time,noise_time_))
        
        for i in range(len(noise_time_)):
            # Step 5.2.1 :Noise interpolation 
            delta_t         = (noise_time[i] -time[cpt]) / (time[cpt+1] - time[cpt])
            SPL_lower       = conditions.noise.hemisphere_SPL_dBA[cpt].reshape(len(phi),len(theta))
            SPL_uppper      = conditions.noise.hemisphere_SPL_dBA[cpt+1].reshape(len(phi),len(theta))
            SPL_gradient    = SPL_uppper -  SPL_lower
            SPL_interp      = SPL_lower + SPL_gradient *delta_t
     
            #  Step 5.2.2 Create surrogate   
            SPL_dBA_surrogate = RegularGridInterpolator((phi, theta),SPL_interp  ,method = 'linear',   bounds_error=False, fill_value=None)       
            
            #  Step 5.2.3 Query surrogate
            R                 = np.linalg.norm(RML[i], axis=1) 
            locs              = np.argsort(R)[:n]
            pts               = (PHI[i][locs],THETA[i][locs]) 
            SPL_dBA_unscaled  = SPL_dBA_surrogate(pts) 
            
            #  Step 5.2.4 Scale data using radius  
            R_ref                = settings.noise_hemisphere_radius  
            SPL_dBA_scaled       = SPL_dBA_unscaled - 20*np.log10(R[locs]/R_ref)
            
            SPL_dBA_temp         = SPL_dBA[idx].flatten()
            SPL_dBA_temp[locs]   = SPL_dBA_scaled
            SPL_dBA[idx]         = SPL_dBA_temp.reshape(N_gm_x,N_gm_y) 
            mic_locs[idx]        = locs 
            idx += 1
            
            if noise_time[i] >= time[cpt+1]:
                cpt += 1             
                
    # Step 6: Make any readings less that background noise equal to background noise
    SPL_dBA                             = np.nan_to_num(SPL_dBA) 
    SPL_dBA[SPL_dBA<background_noise()] = background_noise()  
     
    # Step 7: Store data 
    noise_data.SPL_dBA               = SPL_dBA
    noise_data.time                  = Time 
    noise_data.aircraft_position     = Aircraft_pos
    noise_data.microhpone_locations  = mic_locs
    
    # Step 8: Perform noise metric calculations
    if evalaute_noise_metrics:
        compute_noise_metrics(noise_data, flight_times)
    
    return noise_data


# RCAIDE/Library/Methods/Propulsors/Common/append_avionics_conditions.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke  

from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  append_motor_conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_avionics_conditions(avionics,segment,bus):  
    ones_row    = segment.state.ones_row
    segment.state.conditions.energy[bus.tag][avionics.tag]            = Conditions()
    segment.state.conditions.energy[bus.tag][avionics.tag].power      = 0 * ones_row(1) 
    
    return 



# RCAIDE/Library/Plots/Energy/plot_battery_degradation.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots.Common import set_axes, plot_style 
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------   
def plot_battery_degradation(results,
                            save_figure = False,
                            line_color = 'bo-',
                            line_color2 = 'rs--',
                            save_filename = "Battery_Degradation",
                            file_type = ".png",
                            width = 11, height = 7):
    """This plots the solar flux and power train performance of an solar powered aircraft

    Assumptions:
    None
    
    Source:
    None    
    
    Inputs:
    results.segments.conditions.propulsion
        solar_flux
        battery_power_draw
        battery_energy
    
    Outputs:
    Plots
    
    Properties Used:
    N/A
    """ 
    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters)
    

    for network in results.segments[0].analyses.energy.vehicle.networks: 
        busses  = network.busses
        for bus in busses:
            if bus.identical_battery_modules:
                for i, battery in enumerate(bus.battery_modules):
                    if i == 0:
                        fig = plt.figure(save_filename + '_' + battery.tag)
                        fig.set_size_inches(width,height)  
                        num_segs          = len(results.segments)
                        time_hrs          = np.zeros(num_segs)  
                        capacity_fade     = np.zeros_like(time_hrs)
                        resistance_growth = np.zeros_like(time_hrs)
                        cycle_day         = np.zeros_like(time_hrs)
                        charge_throughput = np.zeros_like(time_hrs)    
                             
                        for i in range(len(results.segments)): 
                            time_hrs[i]    = results.segments[i].conditions.frames.inertial.time[-1,0]  / Units.hour   
                            battery_conditions  = results.segments[i].conditions.energy[bus.tag].battery_modules[battery.tag]    
                            cycle_day[i]          = battery_conditions.cell.cycle_in_day
                            capacity_fade[i]      = battery_conditions.cell.capacity_fade_factor
                            resistance_growth[i]  = battery_conditions.cell.resistance_growth_factor
                            charge_throughput[i]  = battery_conditions.cell.charge_throughput[-1,0]  
                 
                        axis_1 = plt.subplot(3,2,1)
                        axis_1.plot(charge_throughput, capacity_fade, color = ps.color , marker = ps.markers[0], linewidth = ps.line_width ) 
                        axis_1.set_ylabel('$E/E_0$')
                        axis_1.set_xlabel('Ah')
                        set_axes(axis_1)      
                    
                        axis_2 = plt.subplot(3,2,3)
                        axis_2.plot(time_hrs, capacity_fade, color = ps.color, marker = ps.markers[0], linewidth = ps.line_width ) 
                        axis_2.set_ylabel('$E/E_0$')
                        axis_2.set_xlabel('Time (hrs)')
                        set_axes(axis_2)     
                    
                        axis_3 = plt.subplot(3,2,5)
                        axis_3.plot(cycle_day, capacity_fade, color = ps.color, marker = ps.markers[0], linewidth = ps.line_width ) 
                        axis_3.set_ylabel('$E/E_0$')
                        axis_3.set_xlabel('Time (days)')
                        set_axes(axis_3)     
                    
                        axis_4 = plt.subplot(3,2,2) 
                        axis_4.plot(charge_throughput, resistance_growth, color = ps.color, marker = ps.markers[0], linewidth = ps.line_width )
                        axis_4.set_ylabel('$R/R_0$')
                        axis_4.set_xlabel('Ah')
                        set_axes(axis_4)      
                    
                        axis_5 = plt.subplot(3,2,4) 
                        axis_5.plot(time_hrs, resistance_growth, color = ps.color, marker = ps.markers[0], linewidth = ps.line_width )
                        axis_5.set_ylabel('$R/R_0$')
                        axis_5.set_xlabel('Time (hrs)')
                        set_axes(axis_5)     
                    
                        axis_6 = plt.subplot(3,2,6) 
                        axis_6.plot(cycle_day, resistance_growth, color = ps.color, marker = ps.markers[0], linewidth = ps.line_width )
                        axis_6.set_ylabel('$R/R_0$')
                        axis_6.set_xlabel('Time (days)')
                        set_axes(axis_6)              
    
                        
    # set title of plot 
    title_text    = 'Battery Cell Degradation: ' + battery.tag   
    fig.suptitle(title_text) 
    
    plt.tight_layout()    
    if save_figure:    
        fig.savefig(save_filename + '_'+ battery.tag + file_type) 

    return fig 




# RCAIDE/Framework/Analyses/Mission/Segments/Transition/Constant_Acceleration_Constant_Angle_Linear_Climb.py
# 
# 
# Created:  Jul 2023, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports 
from RCAIDE.Framework.Core                            import Units 
from RCAIDE.Framework.Mission.Segments.Evaluate       import Evaluate
from RCAIDE.Library.Mission                           import Common,Segments

# ----------------------------------------------------------------------------------------------------------------------
#  Constant_Acceleration_Constant_Angle_Linear_Climb
# ----------------------------------------------------------------------------------------------------------------------
class Constant_Acceleration_Constant_Angle_Linear_Climb(Evaluate):
    """ Vehicle accelerates at a constant rate between two airspeeds.
    
        Assumptions:
        None
        
        Source:
        None
    """     
    
    def __defaults__(self):
        """ This sets the default solver flow. Anything in here can be modified after initializing a segment.
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            None
    
            Outputs:
            None
    
            Properties Used:
            None
        """           
        
        # --------------------------------------------------------------
        #   User Inputs
        # --------------------------------------------------------------
        self.altitude_start         = None
        self.altitude_end           = None
        self.air_speed_start        = None
        self.climb_angle            = 0.0 * Units['rad'] 
        self.acceleration           = 1.  * Units['m/s/s'] 
        self.pitch_initial          = None
        self.pitch_final            = 0.0 * Units['rad']
        self.true_course            = 0.0 * Units.degrees  
        
        # -------------------------------------------------------------------------------------------------------------- 
        #  Mission specific processes 
        # --------------------------------------------------------------------------------------------------------------  
        initialize                         = self.process.initialize 
        initialize.conditions              = Segments.Transition.Constant_Acceleration_Constant_Angle_Linear_Climb.initialize_conditions  
        iterate                            = self.process.iterate  
        iterate.unknowns.mission           = Common.Unpack_Unknowns.orientation
        iterate.unknowns.controls          = Common.Unpack_Unknowns.control_surfaces
        iterate.residuals.flight_dynamics  = Common.Residuals.flight_dynamics
        
        return




# RCAIDE/Components/Landing_Gear/__init__.py

"""
Module containing landing gear components for aircraft design and analysis. This module 
provides classes for both main and nose landing gear configurations.

See Also
--------
RCAIDE.Library.Components.Fuselages
    Related module for fuselage components that interface with landing gear
RCAIDE.Library.Components.Airfoils
    Related module for airfoil definitions that may affect gear bay design
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .Landing_Gear import Landing_Gear
from .Main_Landing_Gear import Main_Landing_Gear
from .Nose_Landing_Gear import Nose_Landing_Gear


# RCAIDE/Library/Methods/Thermal_Management/Heat_Exchangers/Cross_Flow_Heat_Exchanger/cross_flow_hex_rating_model.py
# 
# 
# Created:  Apr 2024, S. Shekar
# ----------------------------------------------------------------------
#  Imports
# ---------------------------------------------------------------------- 
import numpy as np  

# ----------------------------------------------------------------------
#  Methods
# ----------------------------------------------------------------------
def cross_flow_hex_rating_model(HEX,state,bus,coolant_line, delta_t,t_idx):
    """ Computes the net heat removed by a cross flow heat exchanger and
        resultant coolant and air temperatures. 
          
          Inputs: 
          HEX.
              (all optimized and default properties)
          state.
               conditions.energy.coolant_line.
                                              reservoir.inlet_coolant_temperature [Kelvin]
          Outputs:
               None
               
          Assumptions:
               None
            
          Source:
            Shah RK, Sekulić DP. Fundamentals of Heat Exchanger Design. John Wiley & Sons; 2003 
    """      
    
    air              = HEX.air
    coolant          = HEX.coolant 
    
    # take in variables from the output of sizing problem
    H     = HEX.stack_height
    L_c   = HEX.stack_length
    L_h   = HEX.stack_width 
        
    # Inital assumed efficiency of HEX 
    eff_hex     = 0.75 
    
    # Hydraulic Diameters
    d_h_c       = HEX.coolant_hydraulic_diameter 
    d_h_h       = HEX.air_hydraulic_diameter 
    
    # Fin Height/Spaceing 
    b_c         = HEX.fin_spacing_cold                                       
    b_h         = HEX.fin_spacing_hot
    
    # Fin metal thickness
    delta_h     = HEX.fin_metal_thickness_hot  
    delta_c     = HEX.fin_metal_thickness_cold 
    
    # Platethickness
    delta_w     = HEX.t_w   
    
    # Strip edge exposed 
    l_s_h       = HEX.fin_exposed_strip_edge_hot  
    l_s_c       = HEX.fin_exposed_strip_edge_cold 
    
    #Fin and wall Conductivity 
    fin_conductivity          =HEX.fin_conductivity   
    wall_conductivity         =HEX.wall_conductivity 
                 
    # Ratio of finned area to total area 
    Af_A_h      = HEX.finned_area_to_total_area_hot  
    Af_A_c      = HEX.finned_area_to_total_area_cold        
    
    # Finned area density 
    beta_h      = HEX.fin_area_density_hot  
    beta_c      = HEX.fin_area_density_cold   
    
    # Assumes N passages for hot air and N+1 fpr cold air 
    N_p   = (H-b_c+2*delta_w)/(b_h+b_c+2*delta_w)
    
    # Frontal areas on hot and cokd sides 
    A_fr_h = L_c*H
    A_fr_c = L_h*H
    
    # Heat exchnager volume between plates on each fluid side.  
    V_p_h       = L_h*L_c*b_h*N_p
    V_p_c       = L_h*L_c*b_c*(N_p+1)
    
    # The heat transfer areas 
    A_h         = beta_h*V_p_h
    A_c         = beta_c*V_p_c
    
    #The minimum free flow area 
    A_o_h       = d_h_h*A_h/(4*L_h)     
    A_o_c       = d_h_c*A_c/(4*L_c)     
    
    # Minimum Free flow area  
    sigma_h    = A_o_h/A_fr_h
    sigma_c    = A_o_c/A_fr_c 

    # Inlet temperatures 
    for reservoir in  coolant_line.reservoirs:
        T_i_h  = state.conditions.energy[coolant_line.tag][reservoir.tag].coolant_temperature[t_idx,0] 
    T_i_c           = state.conditions.freestream.temperature[t_idx,0]     
    #turndown_ratio  = battery_conditions.thermal_management_system.HEX.percent_operation[t_idx,0]
    m_dot_h         = HEX.design_coolant_mass_flow_rate#*turndown_ratio
    m_dot_c         = HEX.design_air_mass_flow_rate#*turndown_ratio 
    P_i_c           = HEX.design_air_inlet_pressure#*turndown_ratio
    P_i_h           = HEX.design_coolant_inlet_pressure#*turndown_ratio 
    rho_c_i         = air.compute_density(T_i_h,P_i_c)
    rho_h_i         = coolant.compute_density(T_i_h)  
     
    # Thermal Performance Calculation  
    T_o_h       = T_i_h-eff_hex*(T_i_h-T_i_c)
    T_o_c       = T_i_c+eff_hex*(m_dot_h/m_dot_c)*(T_i_h-T_i_c)
    
    check             = 0
    itetation_counter = 0
    
    while check==0: 
        T_m_h       = (T_i_h+T_o_h)/2
        T_m_c       = (T_i_c+T_o_c)/2
    
        #Prandtl Number 
        Pr_h    =  coolant.compute_prandtl_number(T_m_h)
        Pr_c    =  air.compute_prandtl_number(T_m_c)    
    
        #Absolute viscosity 
        mu_h    = coolant.compute_absolute_viscosity(T_m_h)
        mu_c    = air.compute_absolute_viscosity(T_m_c)       
    
        #Specific heat 
        c_p_h   = coolant.compute_cp(T_m_h)/1000 #KJ/kg-K
        c_p_c   = air.compute_cp(T_m_c)/1000     #KJ/kg-K   
    
        # Core mass velcoity  
        G_h   = m_dot_h/A_o_h
        G_c   = m_dot_c/A_o_c
    
        # Calculate Reynolds Number
        Re_h       = G_h * d_h_h / mu_h
        Re_c       = G_c * d_h_c / mu_c
    
        # Calculate the colburn factor and friction factor using curve fitted values (What about for Turbulent regim, check in london and Kays )
        j_c            = 0.0131 * (Re_c / 1000)**(-0.415)
        j_h            = 0.0131 * (Re_h / 1000)**(-0.415)
    
        f_c            = 0.0514 * (Re_c / 1000)**(-0.471)
        f_h            = 0.0514 * (Re_h / 1000)**(-0.471)
    
        j_f_h          = j_h/f_h
        j_f_c          = j_c/f_c
    
        # Heat Transfer Coefficients
        h_h = j_h * G_h * c_p_h / (Pr_h**(2/3))
        h_c = j_c * G_c * c_p_c / (Pr_c**(2/3))
    
        m_f_h = (np.sqrt((2*h_h)/(fin_conductivity*delta_h)))*np.sqrt(1+(delta_h/l_s_h))
        m_f_c = (np.sqrt((2*h_c)/(fin_conductivity*delta_c)))*np.sqrt(1+(delta_c/l_s_c))
    
    
        l_f_h = b_h / 2 - delta_h
        l_f_c = b_c / 2 - delta_c
    
        # Fin Efficiency
        eta_f_h = np.tanh(m_f_h * l_f_h) / (m_f_h * l_f_h)
        eta_f_c = np.tanh(m_f_c * l_f_c) / (m_f_c * l_f_c)
    
        # Overall Efficiency 
        eta_o_h = 1 - (1 - eta_f_h) * Af_A_h
        eta_o_c = 1 - (1 - eta_f_c) * Af_A_c
    
        # Wall ressistance 
        A_w   = L_c*L_h*(2*N_p+2)
        R_w   = delta_w/(wall_conductivity*A_w)
    
        # Calculate overall heat transfer without fouling
        UA    = 1 / ((1 / (eta_o_h * h_h*A_h)) +R_w+ (1 / (eta_o_c* h_c*A_c)))
    
        # Heat Capcity 
        C_h            = m_dot_h*c_p_h
        C_c            = m_dot_c*c_p_c
    
        C_min, C_max   = min(C_h, C_c), max(C_h, C_c)
        C_r            = C_min / C_max  
    
        # NTU 
        NTU            = UA/C_min
    
        # Updated effectiveness and we neglect longitudnal conduction for now 
        eff_hex_updated= (1 - np.exp(((NTU**0.22)/C_r)*(np.exp(-C_r*(NTU**(0.78))) - 1 ))) 
    
        # Heat trannsfer rate
        q             = eff_hex*(T_i_h-T_i_c)*C_min
    
        # Updated Outlet temperatures 
        T_o_h_updated = T_i_h-(q/C_h) 
        T_o_c_updated = T_i_c+(q/C_c) 
        
        residual = [abs(T_o_c-T_o_c_updated),abs(T_o_h-T_o_h_updated),abs(eff_hex-eff_hex_updated)]
    
        if all(element < 0.01 for element in residual) or itetation_counter>=10:
            check = 1 
            eff_hex = eff_hex_updated
        else:
            itetation_counter +=1 
            T_o_c   = T_o_c_updated
            T_o_h   = T_o_h_updated
            eff_hex = eff_hex_updated
    
    
    # ---------------------------------------------------------------------------------------------------------- 
    # Pressure Drop Calculation 
    # ----------------------------------------------------------------------------------------------------------       
    
    # inital assumption 
    P_o_h   = P_i_h 
    P_o_c   = P_i_c  
    
    iteraion_counter_1 = 0
    check              = 0
    delta_p_c          = []
    delta_p_h          = []
    
    while check == 0:
    
        # from the inlet and outlet pressures given the mean density is calcualted. 
        rho_h_o  =  coolant.compute_density(T_o_h)
        rho_c_o  =  air.compute_density(T_o_c,P_o_c)    
        rho_h_m  = 2 / (1 / rho_h_i + 1 / rho_h_o)
        rho_c_m  = 2 / (1 / rho_c_i + 1 / rho_c_o)             
    
        # Kc_c, Ke_c     = compute_heat_exhanger_factors(kc_vals,ke_vals,sigma_c, Re_c) SAI 
        # Need to check if the values obtained from the function are close to what is obtained ere 
        k_c_c = 0.36
        k_c_h = 0.36
        k_e_c = 0.42
        k_e_h = 0.42
    
        # Thermal Resistance on the hot and cold fluid sides
    
        R_h = 1 / (eta_o_h * h_h * A_h)
        R_c = 1 / (eta_o_c * h_c * A_c)
    
        # Compute Wall temperature
        T_w = (T_m_h + (R_h / R_c) * T_m_c) / (1 + R_h / R_c)
    
        # Considering temperature at wall effecting f value of 0.81 changes
        f_h_wall = f_h * np.power(((T_w + 273) / (273 + T_m_h)), 0.81)
        f_c_wall = f_c * np.power(((T_w + 273) / (273 + T_m_c)), 1)
    
        # Calculate Pressure Drop
    
        delta_p_c.append(np.power(G_c, 2) / (2 * rho_c_i) * ((1 - np.power(sigma_c, 2) + k_c_c)
                                                           + 2 * (rho_c_i / rho_c_o - 1) + f_c_wall * 4 * L_c / d_h_c *
                                                                   rho_c_i / rho_c_m
                                                                 - (1 - np.power(sigma_c, 2) - k_e_c) * rho_c_i / rho_c_o))
    
        delta_p_h.append(np.power(G_h, 2) / (2 * rho_h_i) * ((1 - np.power(sigma_h, 2) + k_c_h)
                                                           + 2 * (rho_h_i / rho_h_o - 1) + f_h_wall * 4 * L_h / d_h_h *
                                                                   rho_h_i / rho_h_m
                                                                    - (1 - np.power(sigma_h, 2) - k_e_h) * rho_h_i / rho_h_o))   
    
        if iteraion_counter_1 >=1:
            residual_pressure    = [abs(delta_p_c[iteraion_counter_1]-delta_p_c[iteraion_counter_1-1]),abs(delta_p_h[iteraion_counter_1]-delta_p_h[iteraion_counter_1-1])]
    
            if all(element < 0.01 for element in residual_pressure) or iteraion_counter_1 >= 10:
                check =1
            else:
                P_o_c                         = (-delta_p_c[iteraion_counter_1]+P_i_c)
                P_o_h                         = (-delta_p_h[iteraion_counter_1]+P_i_h)
                iteraion_counter_1           += 1
        else:
            P_o_c                         = (-delta_p_c[iteraion_counter_1]+P_i_c)
            P_o_h                         = (-delta_p_h[iteraion_counter_1]+P_i_h)
            iteraion_counter_1           += 1
    
    state.conditions.energy[coolant_line.tag][HEX.tag].pressure_diff_air[t_idx+1]          = delta_p_c[iteraion_counter_1]
    state.conditions.energy[coolant_line.tag][HEX.tag].air_mass_flow_rate[t_idx+1]         = m_dot_c  
    
    # Calculate Power drawn by HEX 
    P_coolant = ((m_dot_h*delta_p_h[iteraion_counter_1])/(HEX.pump.efficiency*rho_h_m))
    
    if state.conditions.freestream.velocity[t_idx] > HEX.minimum_air_speed:
        P_air    = 0
    else:
        P_air    = ((m_dot_c*delta_p_c[iteraion_counter_1]/rho_c_m))/HEX.fan.efficiency
    P_hex     = P_air+P_coolant 

    state.conditions.energy[coolant_line.tag][HEX.tag].coolant_mass_flow_rate[t_idx+1]     = m_dot_h  
    state.conditions.energy[coolant_line.tag][HEX.tag].power[t_idx+1]                      = P_hex
    state.conditions.energy[coolant_line.tag][HEX.tag].inlet_air_temperature[t_idx+1]      = T_i_c 
    state.conditions.energy[coolant_line.tag][HEX.tag].outlet_coolant_temperature[t_idx+1] = T_o_h_updated
    state.conditions.energy[coolant_line.tag][HEX.tag].air_mass_flow_rate[t_idx+1]         = m_dot_c  
    state.conditions.energy[coolant_line.tag][HEX.tag].air_inlet_pressure[t_idx+1]         = P_i_c 
    state.conditions.energy[coolant_line.tag][HEX.tag].coolant_inlet_pressure[t_idx+1]     = P_i_h
    state.conditions.energy[coolant_line.tag][HEX.tag].effectiveness_HEX[t_idx+1]          = eff_hex  
    if not state.conditions.energy.recharging: 
        state.conditions.energy[bus.tag].power_draw[t_idx+1]                                  += P_hex 
    
    return  


# RCAIDE/Library/Plots/Performance/Aerodynamics/plot_airfoil_surface_forces.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
 
from RCAIDE.Framework.Core import Units  
import matplotlib.pyplot as plt    

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------     
def plot_airfoil_surface_forces(ap, save_figure = False , arrow_color = 'red',save_filename = 'Airfoil_Cp_Distribution', show_figure = True, file_type = ".png"):  
    """ This plots the forces on an airfoil surface
    
        Assumptions:
        None
        
        Inputs: 
        ap       - data stucture of airfoil boundary layer properties and polars 
         
        Outputs: 
        None 
        
        Properties Used:
        N/A
        """        
    
    # determine dimension of angle of attack and reynolds number 
    n_cpts   = len(ap.AoA)
    nAoA     = len(ap.AoA[0])
    n_pts    = len(ap.x[0,0,:])- 1 
     

    for i in range(n_cpts):     
        for j in range(nAoA): 
            label =  '_AoA_' + str(round(ap.AoA[i][j]/Units.degrees,2)) + '_deg_Re_' + str(round(ap.Re[i][j]/1000000,2)) + 'E6'
            fig   = plt.figure('Airfoil_Pressure_Normals' + label )
            axis = fig.add_subplot(1,1,1) 
            axis.plot(ap.x[0,0,:], ap.y[0,0,:],'k-')   
            for k in range(n_pts):
                dx_val = ap.normals[i,j,k,0]*abs(ap.cp[i,j,k])*0.1
                dy_val = ap.normals[i,j,k,1]*abs(ap.cp[i,j,k])*0.1
                if ap.cp[i,j,k] < 0:
                    plt.arrow(x= ap.x[i,j,k], y=ap.y[i,j,k] , dx= dx_val , dy = dy_val , 
                              fc=arrow_color, ec=arrow_color,head_width=0.005, head_length=0.01 )   
                else:
                    plt.arrow(x= ap.x[i,j,k]+dx_val , y= ap.y[i,j,k]+dy_val , dx= -dx_val , dy = -dy_val , 
                              fc=arrow_color, ec=arrow_color,head_width=0.005, head_length=0.01 )   
    
    
    return fig 




# RCAIDE/Library/Methods/Geometry/Two_Dimensional/Planform/wing_segmented_planform.py
# 
# 
# Created:  Jul 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------    
from RCAIDE.Framework.Core import Data 

# package imports 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  Wing Segmented Planform
# ----------------------------------------------------------------------------------------------------------------------    
def wing_segmented_planform(wing, overwrite_reference = False):
    """Computes standard wing planform values.
    
    Assumptions:
    Multisegmented wing. There is no unexposed wetted area, ie wing area that 
    intersects inside a fuselage. Aerodynamic center is at 25% mean aerodynamic chord.
    
    Source:
    None
    
    Inputs:
    overwrite_reference        <boolean> Determines if reference area, wetted area, and aspect
                                         ratio are overwritten based on the segment values.
    wing.
      chords.root              [m]
      spans.projected          [m]
      symmetric                <boolean> Determines if wing is symmetric
    
    Outputs:
    wing.
      spans.total                [m]
      chords.tip                 [m]
      chords.mean_aerodynamics   [m]
      wing.chords.mean_geometric [m]
      areas.reference            [m^2]
      taper                      [-]
      sweeps.quarter_chord       [radians]
      aspect_ratio               [-]
      thickness_to_chord         [-]
      dihedral                   [radians]
      
      aerodynamic_center         [m]      x, y, and z location

        
    
    Properties Used:
    N/A
    """
    
    # Unpack
    span = wing.spans.projected
    RC   = wing.chords.root
    sym  = wing.symmetric
    
    # Pull all the segment data into array format
    span_locs = []
    twists    = []
    sweeps    = []
    dihedrals = []
    chords    = []
    t_cs      = []
    for key in wing.segments.keys():
        seg = wing.segments[key]
        span_locs.append(seg.percent_span_location)
        twists.append(seg.twist)
        chords.append(seg.root_chord_percent)
        sweeps.append(seg.sweeps.quarter_chord)
        t_cs.append(seg.thickness_to_chord)
        dihedrals.append(seg.dihedral_outboard)
        
    # Convert to arrays
    chords    = np.array(chords)
    span_locs = np.array(span_locs)
    sweeps    = np.array(sweeps)
    t_cs      = np.array(t_cs)
    
    # Basic calcs:
    semispan     = span/(1+sym)
    lengths_ndim = span_locs[1:]-span_locs[:-1]
    lengths_dim  = lengths_ndim*semispan
    chords_dim   = RC*chords
    tapers       = chords[1:]/chords[:-1]
    
    # Calculate the areas of each segment
    As = (lengths_dim*chords_dim[:-1]-(chords_dim[:-1]-chords_dim[1:])*(lengths_dim/2))
    
    # Calculate the weighted area, this should not include any unexposed area 
    A_wets = 2*(1+0.2*t_cs[:-1])*As
    wet_area = np.sum(A_wets)
    
    # Calculate the wing area
    ref_area = np.sum(As)*(1+sym)
    
    # Calculate the Aspect Ratio
    AR = (span**2)/ref_area
    
    # Calculate the total span
    lens = lengths_dim/np.cos(dihedrals[:-1])
    total_len = np.sum(np.array(lens))*(1+sym)
    
    # Calculate the mean geometric chord
    mgc = ref_area/span
    
    # Calculate the mean aerodynamic chord
    A = chords_dim[:-1]
    B = (A-chords_dim[1:])/(-lengths_ndim)
    C = span_locs[:-1]
    integral = ((A+B*(span_locs[1:]-C))**3-(A+B*(span_locs[:-1]-C))**3)/(3*B)
    # For the cases when the wing doesn't taper in a spot
    integral[np.isnan(integral)] = (A[np.isnan(integral)]**2)*((lengths_ndim)[np.isnan(integral)])
    MAC = (semispan*(1+sym)/(ref_area))*np.sum(integral)
    
    # Calculate the taper ratio
    lamda = chords[-1]/chords[0]
    
    # the tip chord
    ct = chords_dim[-1]
    
    # Calculate an average t/c weighted by area
    t_c = np.sum(As*t_cs[:-1])/(ref_area/2)
    
    # Calculate the segment leading edge sweeps
    r_offsets = chords_dim[:-1]/4
    t_offsets = chords_dim[1:]/4
    le_sweeps = np.arctan((r_offsets+np.tan(sweeps[:-1])*(lengths_dim)-t_offsets)/(lengths_dim))    
    
    # Calculate the effective sweeps
    c_4_sweep   = np.arctan(np.sum(lengths_ndim*np.tan(sweeps[:-1])))
    le_sweep_total= np.arctan(np.sum(lengths_ndim*np.tan(le_sweeps)))

    # Calculate the aerodynamic center, but first the centroid
    dxs = np.cumsum(np.concatenate([np.array([0]),np.tan(le_sweeps[:-1])*lengths_dim[:-1]]))
    dys = np.cumsum(np.concatenate([np.array([0]),lengths_dim[:-1]]))
    dzs = np.cumsum(np.concatenate([np.array([0]),np.tan(dihedrals[:-2])*lengths_dim[:-1]]))
    
    Cxys = []
    for i in range(len(lengths_dim)):
        Cxys.append(segment_centroid(le_sweeps[i],lengths_dim[i],dxs[i],dys[i],dzs[i], tapers[i], 
                                     As[i], dihedrals[i], chords_dim[i], chords_dim[i+1]))

    aerodynamic_center = (np.dot(np.transpose(Cxys),As)/(ref_area/(1+sym)))

    single_side_aerodynamic_center = (np.array(aerodynamic_center)*1.)
    single_side_aerodynamic_center[0] = single_side_aerodynamic_center[0] - MAC*.25    
    if sym== True:
        aerodynamic_center[1] = 0
        
    aerodynamic_center[0] = single_side_aerodynamic_center[0]
    
    # Total length for supersonics
    total_length = np.tan(le_sweep_total)*semispan + chords[-1]*RC
    
    # Pack stuff
    if overwrite_reference:
        wing.areas.reference         = ref_area
        wing.areas.wetted            = wet_area
        wing.aspect_ratio            = AR

    wing.spans.total                    = total_len
    wing.chords.mean_geometric          = mgc
    wing.chords.mean_aerodynamic        = MAC
    wing.chords.tip                     = ct
    wing.taper                          = lamda
    wing.sweeps.quarter_chord           = c_4_sweep
    wing.sweeps.leading_edge            = le_sweep_total
    wing.thickness_to_chord             = t_c
    wing.aerodynamic_center             = aerodynamic_center
    wing.single_side_aerodynamic_center = single_side_aerodynamic_center
    wing.total_length                   = total_length 

    # update remainder segment properties
    segment_properties(wing)
    
    return wing
 
def segment_properties(wing,update_wet_areas=False,update_ref_areas=False):
    """Computes detailed segment properties. These are currently used for parasite drag calculations.

    Assumptions:
    Segments are trapezoids

    Source:
    http://aerodesign.stanford.edu/aircraftdesign/aircraftdesign.html (Stanford AA241 A/B Course Notes)

    Inputs:
    wing.
      exposed_root_chord_offset [m]
      symmetric                 [-]
      spans.projected           [m]
      thickness_to_chord        [-]
      areas.wetted              [m^2]
      chords.root               [m]
      Segments.
        percent_span_location   [-]
        root_chord_percent      [-]

    Outputs:
    wing.areas.wetted           [m^2]
    wing.areas.reference        [m^2]
    wing.segments.
      taper                     [-]
      chords.mean_aerodynamic   [m]
      areas.
        reference               [m^2]
        exposed                 [m^2]
        wetted                  [m^2]
        

    Properties Used:
    N/A
    """  
        
    # Unpack wing
    exposed_root_chord_offset = wing.exposed_root_chord_offset
    symm                      = wing.symmetric
    semispan                  = wing.spans.projected*0.5 * (2 - symm)
    t_c_w                     = wing.thickness_to_chord
    segments                  = wing.segments
    segment_names             = list(segments.keys())
    num_segments              = len(segment_names)   
    total_wetted_area         = 0.
    total_reference_area      = 0.
    root_chord                = wing.chords.root      
    
    for i_segs in range(num_segments):
        if i_segs == num_segments-1:
            continue 
        else:  
            span_seg  = semispan*(segments[segment_names[i_segs+1]].percent_span_location - segments[segment_names[i_segs]].percent_span_location ) 
            segment   = segments[segment_names[i_segs]]         
            
            if i_segs == 0:
                chord_root    = root_chord*segments[segment_names[i_segs]].root_chord_percent
                chord_tip     = root_chord*segments[segment_names[i_segs+1]].root_chord_percent   
                wing_root     = chord_root + exposed_root_chord_offset*((chord_tip - chord_root)/span_seg)
                taper         = chord_tip/wing_root  
                mac_seg       = wing_root  * 2/3 * (( 1 + taper  + taper**2 )/( 1 + taper))  
                Sref_seg      = span_seg*(chord_root+chord_tip)*0.5 
                S_exposed_seg = (span_seg-exposed_root_chord_offset)*(wing_root+chord_tip)*0.5                    
            
            else: 
                chord_root    = root_chord*segments[segment_names[i_segs]].root_chord_percent
                chord_tip     = root_chord*segments[segment_names[i_segs+1]].root_chord_percent
                taper         = chord_tip/chord_root   
                mac_seg       = chord_root * 2/3 * (( 1 + taper  + taper**2 )/( 1 + taper))
                Sref_seg      = span_seg*(chord_root+chord_tip)*0.5
                S_exposed_seg = Sref_seg

            if wing.symmetric:
                Sref_seg = Sref_seg*2
                S_exposed_seg = S_exposed_seg*2
            
            # compute wetted area of segment
            if t_c_w < 0.05:
                Swet_seg = 2.003* S_exposed_seg
            else:
                Swet_seg = (1.977 + 0.52*t_c_w) * S_exposed_seg
                
            segment.taper                   = taper
            segment.chords                  = Data()
            segment.chords.mean_aerodynamic = mac_seg
            segment.areas                   = Data()
            segment.areas.reference         = Sref_seg
            segment.areas.exposed           = S_exposed_seg
            segment.areas.wetted            = Swet_seg
            
            total_wetted_area    = total_wetted_area + Swet_seg
            total_reference_area = total_reference_area + Sref_seg 
    
    if wing.areas.reference==0. or update_ref_areas:
        wing.areas.reference = total_reference_area
        
    if wing.areas.wetted==0. or update_wet_areas:
        wing.areas.wetted    = total_wetted_area
        
    return wing

# Segment centroid
def segment_centroid(le_sweep,seg_span,dx,dy,dz,taper,A,dihedral,root_chord,tip_chord):
    """Computes the centroid of a trapezoidal segment
    
    Assumptions:
    Polygon
    
    Source:
    None
    
    Inputs:
    le_sweep      [rad]
    seg_span      [m]
    dx            [m]
    dy            [m]
    taper         [dimensionless]
    A             [m**2]
    dihedral      [radians]
    root_chord    [m]
    tip_chord     [m]

    Outputs:
    cx,cy         [m,m]

    Properties Used:
    N/A
    """    
    
    a = tip_chord
    b = root_chord
    c = np.tan(le_sweep)*seg_span
    cx = (2*a*c + a**2 + c*b + a*b + b**2) / (3*(a+b))
    cy = seg_span / 3. * (( 1. + 2. * taper ) / (1. + taper))
    cz = cy * np.tan(dihedral)    
    
    return np.array([cx+dx,cy+dy,cz+dz])




# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

# RCAIDE Imports
from   RCAIDE                    import * 
from   RCAIDE.Library.Components import Wings 

# ----------------------------------------------------------------------
#  Compute asymmetry drag due to engine failure 
# ----------------------------------------------------------------------

## @ingroup Methods-Aerodynamics-Common-Fidelity_Zero-Helper_Functions
def asymmetry_drag(state, geometry, engine_out_location = 0,  single_engine_thrust = 0,  windmilling_drag_coefficient = 0.):
    """Computes asymmetry drag due to engine failure

    Assumptions:
    Two engine aircraft

    Source:
    Unknown source

    Inputs:
    state.conditions.freestream.dynamic_pressure                                                [Pa]
    state.conditions.aerodynamics.drag_breakdown.windmilling_drag.windmilling_drag_coefficient  [Unitless]
    geometry.
      mass_properties.center_of_gravity                                                         [m]
      networks. 
        number_of_engines                                                                       [Unitless]
      wings.
        tag                                                                                     
        sref (this function probably doesn't run)                                               [m^2]
	spans.projected                                                                         [m]
      reference_area                                                                            [m^2]

    Outputs:
    asymm_trim_drag_coefficient                                                                 [Unitless]
    (packed in state.conditions.aerodynamics.drag_breakdown.asymmetry_trim_coefficient)

    Properties Used:
    N/A
    """ 
    # ==============================================
	# Unpack
    # ==============================================
    vehicle    = geometry 
    wings      = vehicle.wings
    dyn_press  = state.conditions.freestream.dynamic_pressure
    
     # Defining reference area
    if vehicle.reference_area:
            reference_area = vehicle.reference_area
    else:
        n_wing = 0
        for wing in wings:
            if not isinstance(wing,Wings.Main_Wing): continue
            n_wing = n_wing + 1
            reference_area = wing.sref
        if n_wing > 1:
            print(' More than one Main_Wing in the vehicle. Last one will be considered.')
        elif n_wing == 0:
            print('No Main_Wing defined! Using the 1st wing found')
            for wing in wings:
                if not isinstance(wing,Wings.Wing): continue
                reference_area = wing.sref
                break
            
    # getting cg x position
    xcg = vehicle.mass_properties.center_of_gravity[0]  
    
    # finding vertical tail
    for idx,wing in enumerate(wings):
        if not wing.vertical: continue
        vertical_idx = wing.tag
        break
    # if vertical tail not found, raise error
    try:
        vertical_idx
    except AttributeError:
        print(' No vertical tail found! Error calculating one engine inoperative drag')

    # getting vertical tail data (span, distance to cg)
    vertical_height = wings[vertical_idx].spans.projected
    vertical_dist   = wings[vertical_idx].aerodynamic_center[0] + wings[vertical_idx].origin[0][0] - xcg[0]
    
    # colculating windmilling drag
    if windmilling_drag_coefficient == 0:
        try:
            windmilling_drag_coefficient = state.conditions.aerodynamics.coefficients.drag.windmilling.total  
        except: pass
    
    windmilling_drag = windmilling_drag_coefficient * dyn_press * reference_area
    
    # calculating Drag force due to trim     
    trim_drag = (engine_out_location**2 * (single_engine_thrust+windmilling_drag)**2 ) /      \
                (dyn_press * 3.141593* (vertical_height*vertical_dist)**2)

    # Compute asymmetry trim drag coefficient
    asymm_trim_drag_coefficient = trim_drag / dyn_press / reference_area

    # dump data to state
    state.conditions.aerodynamics.coefficients.drag.asymmetry_trim.total = asymm_trim_drag_coefficient

    return asymm_trim_drag_coefficient


# compute_wing_induced_velocity.py
# 
# Created:  Dec 2020, E. Botero
# Modified: May 2021, E. Botero  
#           Jun 2021, E. Botero  

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

# package imports 
import numpy as np 

def compute_wing_induced_velocity(VD,mach,compute_EW=False):
    """ This computes the induced velocities at each control point of the vehicle vortex lattice 

    Assumptions: 
    Trailing vortex legs infinity are alligned to freestream
    
    Outside of a call to the VLM() function itself, EW does not need to be computed, as C_mn 
    provides the same information in the body-frame. 

    Source:  
    1. Miranda, Luis R., Robert D. Elliot, and William M. Baker. "A generalized vortex 
    lattice method for subsonic and supersonic flow applications." (1977). (NASA CR)
    
    2. VORLAX Source Code

    Inputs: 
    VD       - vehicle vortex distribution                    [Unitless] 
    mach                                                      [Unitless] 
    
    Outputs:                                
    C_mn     - total induced velocity matrix                  [Unitless] 
    s        - semispan of the horshoe vortex                 [m] 
    t        - tangent of the horshoe vortex                  [-] 
    CHORD    - chord length for a panel                       [m] 
    RFLAG    - sonic vortex flag                              [boolean] 
    ZETA     - tangent incidence angle of the chordwise strip [-] 

    Properties Used:
    N/A
    """
    # unpack  
    LE_ind       = VD.leading_edge_indices
    TE_ind       = VD.trailing_edge_indices
    n_cp         = VD.n_cp
    n_mach       = len(mach)
    mach         = np.array(mach,dtype=np.float32)

    # Control points from the VLM 
    XAH   = np.array(np.atleast_2d(VD.XAH*1.),dtype=np.float32)
    YAH   = np.array(np.atleast_2d(VD.YAH*1.),dtype=np.float32)
    ZAH   = np.array(np.atleast_2d(VD.ZAH*1.),dtype=np.float32)
    XBH   = np.array(np.atleast_2d(VD.XBH*1.),dtype=np.float32)
    YBH   = np.array(np.atleast_2d(VD.YBH*1.),dtype=np.float32)
    ZBH   = np.array(np.atleast_2d(VD.ZBH*1.),dtype=np.float32)
    XA1   = np.array(np.atleast_2d(VD.XA1*1.),dtype=np.float32)
    YA1   = np.array(np.atleast_2d(VD.YA1*1.),dtype=np.float32)
    ZA1   = np.array(np.atleast_2d(VD.ZA1*1.),dtype=np.float32)
    XB1   = np.array(np.atleast_2d(VD.XB1*1.),dtype=np.float32)
    YB1   = np.array(np.atleast_2d(VD.YB1*1.),dtype=np.float32)
    ZB1   = np.array(np.atleast_2d(VD.ZB1*1.),dtype=np.float32)
    XA2   = np.array(np.atleast_2d(VD.XA2*1.),dtype=np.float32)
    YA2   = np.array(np.atleast_2d(VD.YA2*1.),dtype=np.float32)
    ZA2   = np.array(np.atleast_2d(VD.ZA2*1.),dtype=np.float32)
    XB2   = np.array(np.atleast_2d(VD.XB2*1.),dtype=np.float32)
    YB2   = np.array(np.atleast_2d(VD.YB2*1.),dtype=np.float32)
    ZB2   = np.array(np.atleast_2d(VD.ZB2*1.),dtype=np.float32)
    XC    = np.array(np.atleast_2d(VD.XC*1.),dtype=np.float32)
    YC    = np.array(np.atleast_2d(VD.YC*1.),dtype=np.float32)
    ZC    = np.array(np.atleast_2d(VD.ZC*1.),dtype=np.float32)
    XA_TE = np.array(np.atleast_2d(VD.XA_TE*1.),dtype=np.float32)
    XB_TE = np.array(np.atleast_2d(VD.XB_TE*1.),dtype=np.float32)
    
    
    # Panel Dihedral Angle, using AH and BH location
    D      = np.sqrt((YAH-YBH)**2+(ZAH-ZBH)**2)
    COS_DL = (YBH-YAH)/D    
    DL     = np.arccos(COS_DL)
    DL[DL>np.pi/2] = DL[DL>np.pi/2] - np.pi # This flips the dihedral angle for the other side of the wing
    
    # -------------------------------------------------------------------------------------------
    # Compute velocity induced by horseshoe vortex segments on every control point by every panel
    # ------------------------------------------------------------------------------------------- 
    # If YBH is negative, flip A and B, ie negative side of the airplane. Vortex order flips
    boolean = YAH>YBH
    XA1[boolean], XB1[boolean] = XB1[boolean], XA1[boolean]
    YA1[boolean], YB1[boolean] = YB1[boolean], YA1[boolean]
    ZA1[boolean], ZB1[boolean] = ZB1[boolean], ZA1[boolean]
    XA2[boolean], XB2[boolean] = XB2[boolean], XA2[boolean]
    YA2[boolean], YB2[boolean] = YB2[boolean], YA2[boolean]
    ZA2[boolean], ZB2[boolean] = ZB2[boolean], ZA2[boolean]    
    XAH[boolean], XBH[boolean] = XBH[boolean], XAH[boolean]
    YAH[boolean], YBH[boolean] = YBH[boolean], YAH[boolean] 
    ZAH[boolean], ZBH[boolean] = ZBH[boolean], ZAH[boolean]

    XA_TE[boolean], XB_TE[boolean] = XB_TE[boolean], XA_TE[boolean]
    
    # These vortices will use AH and BH, rather than the typical location
    xa = XAH
    ya = YAH
    za = ZAH
    xb = XBH
    yb = YBH
    zb = ZBH
    
    # This is not the control point for the panel, its the middle front of the vortex
    xc = 0.5*(xa+xb)
    yc = 0.5*(ya+yb)
    zc = 0.5*(za+zb)
    
    # This is the receiving point, or the control points
    xo = XC.T
    yo = YC.T
    zo = ZC.T
    
    # Incline the vortex
    theta    = np.arctan2(zb-za,yb-ya)
    costheta = np.cos(theta)
    sintheta = np.sin(theta)
    
    # rotated axes
    x1bar = (xb - xc)
    y1bar = (yb - yc)*costheta + (zb - zc)*sintheta
    
    xobar = (xo - xc)
    yobar = (yo - yc)*costheta + (zo - zc)*sintheta
    zobar =-(yo - yc)*sintheta + (zo - zc)*costheta
    
    # COMPUTE COORDINATES OF RECEIVING POINT WITH RESPECT TO END POINTS OF SKEWED LEG.
    shape   = np.shape(xobar)
    shape_0 = shape[0]
    shape_1 = shape[1]
    s       = np.abs(y1bar)
    t       = x1bar/y1bar  
    s       = np.repeat(s,shape_0,axis=0)
    t       = np.repeat(t,shape_0,axis=0)
    
    X1 = xobar + t*s # In a planar case XC-XAH
    Y1 = yobar + s   # In a planar case YC-YAH
    X2 = xobar - t*s # In a planar case XC-XBH
    Y2 = yobar - s   # In a planar case YC-YBH
    
    # The cutoff hardcoded into vorlax
    CUTOFF = 0.8
    
    # CALCULATE AXIAL DISTANCE BETWEEN PROJECTION OF RECEIVING POINT ONTO HORSESHOE PLANE AND EXTENSION OF SKEWED LEG.
    XTY = xobar - t*yobar
    
    # The notation in this method is flipped from the paper
    B2 = np.atleast_3d(mach**2-1.)
    
    # SET VALUES OF NUMERICAL TOLERANCE CONSTANTS.
    TOL    = s /500.0
    TOLSQ  = TOL *TOL
    TOLSQ2 = 2500.0 *TOLSQ
    ZSQ    = zobar *zobar
    YSQ1   = Y1 *Y1
    YSQ2   = Y2 *Y2
    RTV1   = YSQ1 + ZSQ
    RTV2   = YSQ2 + ZSQ
    XSQ1   = X1 *X1
    XSQ2   = X2 *X2
    
    # Split the vectors into subsonic and supersonic
    sub      = (B2<0)[:,0,0]
    B2_sub   = B2[sub,:,:]
    RO1_sub  = B2_sub*RTV1
    RO2_sub  = B2_sub*RTV2
    
    # ZERO-OUT PERTURBATION VELOCITY COMPONENTS
    U = np.zeros((n_mach,shape_0,shape_1),dtype=np.float32)
    V = np.zeros((n_mach,shape_0,shape_1),dtype=np.float32)
    W = np.zeros((n_mach,shape_0,shape_1),dtype=np.float32)    
    
    if np.sum(sub)>0:
        # COMPUTATION FOR SUBSONIC HORSESHOE VORTEX
        U[sub], V[sub], W[sub] = subsonic(zobar,XSQ1,RO1_sub,XSQ2,RO2_sub,XTY,t,B2_sub,ZSQ,TOLSQ,X1,Y1,X2,Y2,RTV1,RTV2)   

    
    # COMPUTATION FOR SUPERSONIC HORSESHOE VORTEX. some values computed in a preprocessing section in VLM
    sup         = (B2>=0)[:,0,0]
    B2_sup      = B2[sup,:,:]
    RO1_sup     = B2[sup,:,:]*RTV1
    RO2_sup     = B2[sup,:,:]*RTV2
    RNMAX       = VD.panels_per_strip
    CHORD       = VD.chord_lengths
    CHORD       = np.repeat(CHORD,shape_0,axis=0)
    RFLAG       = np.ones((n_mach,shape_1),dtype=np.int8)
    
    if np.sum(sup)>0:
        U[sup], V[sup], W[sup], RFLAG[sup,:] = supersonic(zobar,XSQ1,RO1_sup,XSQ2,RO2_sup,XTY,t,B2_sup,ZSQ,TOLSQ,TOL,TOLSQ2,\
                                                    X1,Y1,X2,Y2,RTV1,RTV2,CUTOFF,CHORD,RNMAX,n_cp,TE_ind,LE_ind)
         
    
    # Rotate into the vehicle frame and pack into a velocity matrix
    C_mn = np.stack([U, V*costheta - W*sintheta, V*sintheta + W*costheta],axis=-1)
    
    
    if compute_EW == True:
        # Calculate the W velocity in the VORLAX frame for later calcs
        # The angles are Dihedral angle of the current panel - dihedral angle of the influencing panel
        COS1   = np.cos(DL.T - DL)
        SIN1   = np.sin(DL.T - DL) 
        WEIGHT = 1
        
        EW = (W*COS1-V*SIN1)*WEIGHT
    else:
        # Assume that this function is being used outside of VLM, EW is not needed
        EW = np.nan
        

    return C_mn, s, RFLAG, EW
    
def subsonic(Z,XSQ1,RO1,XSQ2,RO2,XTY,T,B2,ZSQ,TOLSQ,X1,Y1,X2,Y2,RTV1,RTV2):
    """  This computes the induced velocities at each control point 
    of the vehicle vortex lattice for subsonic mach numbers

    Assumptions: 
    Trailing vortex legs infinity are alligned to freestream

    Source:  
    1. Miranda, Luis R., Robert D. Elliot, and William M. Baker. "A generalized vortex 
    lattice method for subsonic and supersonic flow applications." (1977). (NASA CR)
    
    2. VORLAX Source Code

    Inputs: 
    Z       Z relative location of the vortices          [m]
    XSQ1    X1 squared                                   [m^2]
    RO1     coefficient                                  [-]
    XSQ2    X2 squared                                   [m^2]
    RO2     coefficient                                  [-]
    XTY     AXIAL DISTANCE BETWEEN PROJECTION OF RECEIVING POINT ONTO HORSESHOE PLANE AND EXTENSION OF SKEWED LEG [m]
    T       tangent of the horshoe vortex                [-] 
    B2      mach^2-1 (-beta2)                            [-] 
    ZSQ     Z squared                                    [m^2] 
    TOLSQ   coefficient                                  [-]
    X1      X coordinate of the left side of the vortex  [m]
    Y1      Y coordinate of the left side of the vortex  [m]
    X2      X coordinate of the right side of the vortex [m]
    Y2      Y coordinate of the right side of the vortex [m]
    RTV1    coefficient                                  [-]
    RTV2    coefficient                                  [-]

    
    Outputs:           
    U       X velocity       [unitless]
    V       Y velocity       [unitless]
    W       Z velocity       [unitless]

    Properties Used:
    N/A
    """  
    
    CPI  = 4 * np.pi
    RAD1 = np.sqrt(XSQ1 - RO1)
    RAD2 = np.sqrt(XSQ2 - RO2)
    
    TBZ  = (T*T-B2)*ZSQ
    DENOM = XTY * XTY + TBZ
    
    TOLSQ = np.broadcast_to(TOLSQ,np.shape(DENOM))
    
    DENOM[DENOM<TOLSQ] = TOLSQ[DENOM<TOLSQ]
    
    FB1 = (T *X1 - B2 *Y1) /RAD1
    FT1 = (X1 + RAD1) /(RAD1 *RTV1)
    FT1[RTV1<TOLSQ] = 0.
    
    FB2 = (T *X2 - B2 *Y2) /RAD2
    FT2 = (X2 + RAD2) /(RAD2 *RTV2)
    FT2[RTV2<TOLSQ] = 0.
    
    QB = (FB1 - FB2) /DENOM
    ZETAPI = Z /CPI
    U = ZETAPI *QB
    U[ZSQ<TOLSQ] = 0.
    V = ZETAPI * (FT1 - FT2 - QB *T)
    V[ZSQ<TOLSQ] = 0.
    W = - (QB *XTY + FT1 *Y1 - FT2 *Y2) /CPI
    
    return U, V, W

def supersonic(Z,XSQ1,RO1,XSQ2,RO2,XTY,T,B2,ZSQ,TOLSQ,TOL,TOLSQ2,X1,Y1,X2,Y2,RTV1,RTV2,CUTOFF,CHORD,RNMAX,n_cp,TE_ind, LE_ind):
    """  This computes the induced velocities at each control point 
    of the vehicle vortex lattice for supersonic mach numbers

    Assumptions: 
    Trailing vortex legs infinity are alligned to freestream

    Source:  
    1. Miranda, Luis R., Robert D. Elliot, and William M. Baker. "A generalized vortex 
    lattice method for subsonic and supersonic flow applications." (1977). (NASA CR)
    
    2. VORLAX Source Code

    Inputs: 
    Z            Z relative location of the vortices          [m]
    XSQ1         X1 squared                                   [m^2]
    RO1          coefficient                                  [-]
    XSQ2         X2 squared                                   [m^2]
    RO2          coefficient                                  [-]
    XTY          AXIAL DISTANCE BETWEEN PROJECTION OF RECEIVING POINT ONTO HORSESHOE PLANE AND EXTENSION OF SKEWED LEG [m]
    T            tangent of the horshoe vortex                [-] 
    B2           mach^2-1 (-beta2)                            [-] 
    ZSQ          Z squared                                    [m^2] 
    TOLSQ        coefficient                                  [-]
    X1           X coordinate of the left side of the vortex  [m]
    Y1           Y coordinate of the left side of the vortex  [m]
    X2           X coordinate of the right side of the vortex [m]
    Y2           Y coordinate of the right side of the vortex [m]
    RTV1         coefficient                                  [-]
    RTV2         coefficient                                  [-]
    CUTOFF       coefficient                                  [-]
    CHORD        chord length for a panel                     [m] 
    RNMAX        number of chordwise panels                   [-]
    n_cp         number of control points                     [-]
    TE_ind       indices of the trailing edge                 [-]
    LE_ind       indices of the leading edge                  [-]
    

    
    Outputs:           
    U       X velocity        [unitless]
    V       Y velocity        [unitless]
    W       Z velocity        [unitless]
    RFLAG   sonic vortex flag [boolean] 

    Properties Used:
    N/A
    """      
    
    CPI    = 2 * np.pi
    T2     = T*T
    ZETAPI = Z/CPI
    shape  = np.shape(RO1)
    RAD1   = np.sqrt(XSQ1 - RO1)
    RAD2   = np.sqrt(XSQ2 - RO2)
    
    RAD1[np.isnan(RAD1)] = 0. 
    RAD2[np.isnan(RAD2)] = 0. 
    
    DENOM             = XTY * XTY + (T2 - B2) *ZSQ # The last part of this is the TBZ term
    SIGN              = np.ones(shape,dtype=np.int8)
    SIGN[DENOM<0]     = -1.
    TOLSQ             = np.broadcast_to(TOLSQ,shape)
    DENOM_COND        = np.abs(DENOM)<TOLSQ
    DENOM[DENOM_COND] = SIGN[DENOM_COND]*TOLSQ[DENOM_COND]
    
    # Create a boolean for various conditions for F1 that goes to zero
    bool1           = np.ones(shape,dtype=np.bool8)
    bool1[:,X1<TOL] = False
    bool1[RAD1==0.] = False
    RAD1[:,X1<TOL]  = 0.0
    
    REPS = CUTOFF*XSQ1
    FRAD = RAD1

    bool1[RO1>REPS] = False
    FB1 = (T*X1-B2*Y1)/FRAD
    FT1 = X1/(FRAD*RTV1)
    FT1[RTV1<TOLSQ] = 0.
    
    # Use the boolean to turn things off
    FB1[np.isnan(FB1)] = 1.
    FT1[np.isnan(FT1)] = 1.
    FB1[np.isinf(FB1)] = 1.
    FT1[np.isinf(FT1)] = 1.    
    FB1 = FB1*bool1
    FT1 = FT1*bool1
    
    # Round 2
    # Create a boolean for various conditions for F2 that goes to zero
    bool2           = np.ones(shape,dtype=np.bool8)
    bool2[:,X2<TOL] = False
    bool2[RAD2==0.] = False
    RAD2[:,X2<TOL]  = 0.0
    
    REPS = CUTOFF *XSQ2
    FRAD = RAD2    
    
    bool2[RO2>REPS] = False
    FB2 = (T *X2 - B2 *Y2)/FRAD
    FT2 = X2 /(FRAD *RTV2)
    FT2[RTV2<TOLSQ] = 0.
    
    # Use the boolean to turn things off
    FB2[np.isnan(FB2)] = 1.
    FT2[np.isnan(FT2)] = 1.
    FB2[np.isinf(FB2)] = 1.
    FT2[np.isinf(FT2)] = 1.    
    FB2 = FB2*bool2
    FT2 = FT2*bool2
    
    QB = (FB1 - FB2) /DENOM
    U  = ZETAPI *QB
    V  = ZETAPI *(FT1 - FT2 - QB *T)
    W  = - (QB *XTY + FT1 *Y1 - FT2 *Y2) /CPI    
    
    # COMPUTATION FOR SUPERSONIC HORSESHOE VORTEX WHEN RECEIVING POINT IS IN THE PLANE OF THE HORSESHOE
    in_plane = np.broadcast_to(ZSQ<TOLSQ2,shape)
    RAD1_in  = RAD1[in_plane]
    RAD2_in  = RAD2[in_plane] 
    Y1_in    = Y1[ZSQ<TOLSQ2]
    Y2_in    = Y2[ZSQ<TOLSQ2]
    XTY_in   = XTY[ZSQ<TOLSQ2]
    TOL_in   = TOL[ZSQ<TOLSQ2]
    
    if np.sum(in_plane)>0:
        W_in = supersonic_in_plane(RAD1_in, RAD2_in, Y1_in, Y2_in, TOL_in, XTY_in, CPI)
    else:
        W_in = []

    U[in_plane] = 0.
    V[in_plane] = 0.
    W[in_plane] = W_in
    
    # DETERMINE IF TRANSVERSE VORTEX LEG OF HORSESHOE ASSOCIATED TO THE
    # CONTROL POINT UNDER CONSIDERATION IS SONIC (SWEPT PARALLEL TO MACH
    # LINE)? IF SO THEN RFLAG = 0.0, OTHERWISE RFLAG = 1.0.
    size   = shape[1]
    n_mach = shape[0]    
    T2S = np.atleast_2d(T2[0,:])*np.ones((n_mach,1))
    T2F = np.zeros((n_mach,size))
    T2A = np.zeros((n_mach,size))
    
    # Setup masks
    F_mask = np.ones((n_mach,size),dtype=np.bool8)
    A_mask = np.ones((n_mach,size),dtype=np.bool8)
    F_mask[:,TE_ind] = False
    A_mask[:,LE_ind] = False
    
    # Apply the mask
    T2F[A_mask] = T2S[F_mask]
    T2A[F_mask] = T2S[A_mask]
    
    # Zero out terms on the LE and TE
    T2F[:,TE_ind] = 0.
    T2A[:,LE_ind]        = 0.

    TRANS = (B2[:,:,0]-T2F)*(B2[:,:,0]-T2A)
    
    RFLAG = np.ones((n_mach,size),dtype=np.int8)
    RFLAG[TRANS<0] = 0.
    
    FLAG_bool          = np.zeros_like(TRANS,dtype=bool)
    FLAG_bool[TRANS<0] = True
    FLAG_bool          = np.reshape(FLAG_bool,(n_mach,size,-1))
    

    # COMPUTE THE GENERALIZED PRINCIPAL PART OF THE VORTEX-INDUCED VELOCITY INTEGRAL, WWAVE.
    # FROM LINE 2647 VORLAX, the IR .NE. IRR means that we're looking at vortices that affect themselves
    WWAVE   = np.zeros(shape,dtype=np.float32)
    COX     = CHORD /RNMAX
    eye     = np.eye(n_cp,dtype=np.int8)
    T2      = np.broadcast_to(T2,shape)*eye
    B2_full = np.broadcast_to(B2,shape)*eye
    COX     = np.broadcast_to(COX,shape)*eye
    WWAVE[B2_full>T2] = - 0.5 *np.sqrt(B2_full[B2_full>T2] -T2[B2_full>T2] )/COX[B2_full>T2] 

    W = W + WWAVE    
    
    # IF CONTROL POINT BELONGS TO A SONIC HORSESHOE VORTEX, AND THE
    # SENDING ELEMENT IS SUCH HORSESHOE, THEN MODIFY THE NORMALWASH
    # COEFFICIENTS IN SUCH A WAY THAT THE STRENGTH OF THE SONIC VORTEX
    # WILL BE THE AVERAGE OF THE STRENGTHS OF THE HORSESHOES IMMEDIATELY
    # IN FRONT OF AND BEHIND IT.
    
    # Zero out the row
    FLAG_bool_rep     = np.broadcast_to(FLAG_bool,shape)
    W[FLAG_bool_rep]  = 0. # Default to zero

    # The self velocity goes to 2
    FLAG_bool_split   = np.array(np.split(FLAG_bool.ravel(),n_mach))
    FLAG_ind          = np.array(np.where(FLAG_bool_split))
    squares           = np.zeros((size,size,n_mach))
    squares[FLAG_ind[1],FLAG_ind[1],FLAG_ind[0]] = 1
    squares           = np.ravel(squares,order='F')
    
    FLAG_bool_self    = np.where(squares==1)[0]
    W                 = W.ravel()
    W[FLAG_bool_self] = 2. # It's own value, -2
    
    # The panels before and after go to -1
    FLAG_bool_bef = FLAG_bool_self - 1
    FLAG_bool_aft = FLAG_bool_self + 1
    W[FLAG_bool_bef] = -1.
    W[FLAG_bool_aft] = -1.
    
    W = np.reshape(W,shape)

    return U, V, W, RFLAG


def supersonic_in_plane(RAD1,RAD2,Y1,Y2,TOL,XTY,CPI):
    """  This computes the induced velocities at each control point 
    in the special case where the vortices lie in the same plane
    
    Assumptions: 
    Trailing vortex legs infinity are alligned to freestream
    In plane vortices only produce W velocity

    Source:  
    1. Miranda, Luis R., Robert D. Elliot, and William M. Baker. "A generalized vortex 
    lattice method for subsonic and supersonic flow applications." (1977). (NASA CR)
    
    2. VORLAX Source Code

    Inputs: 
    RAD1    array of zeros                               [-]
    RAD2    array of zeros                               [-]
    Y1      Y coordinate of the left side of the vortex  [m]
    Y2      Y coordinate of the right side of the vortex [m]
    TOL     coefficient                                  [-]
    XTY     AXIAL DISTANCE BETWEEN PROJECTION OF RECEIVING POINT ONTO HORSESHOE PLANE AND EXTENSION OF SKEWED LEG [m]
    CPI     2 Pi                                         [radians]

    
    Outputs:           
    W       Z velocity       [unitless]

    Properties Used:
    N/A
    """    
    
    shape = np.shape(RAD2)
    F1    = np.zeros(shape)
    F2    = np.zeros(shape)
    
    reps  = int(shape[0]/np.size(Y1))
    
    Y1  = np.tile(Y1,reps)
    Y2  = np.tile(Y2,reps)
    TOL = np.tile(TOL,reps)
    XTY = np.tile(XTY,reps)
    
    F1[np.abs(Y1)>TOL] = RAD1[np.abs(Y1)>TOL]/Y1[np.abs(Y1)>TOL]
    F2[np.abs(Y2)>TOL] = RAD2[np.abs(Y2)>TOL]/Y2[np.abs(Y2)>TOL]
    
    W = np.zeros(shape)
    W[np.abs(XTY)>TOL] = (-F1[np.abs(XTY)>TOL] + F2[np.abs(XTY)>TOL])/(XTY[np.abs(XTY)>TOL]*CPI)

    return W



# RCAIDE/Library/Components/Energy/Sources/Battery_Modules/Lithium_Air.py
# 
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports
from RCAIDE.Framework.Core import Units
from .Generic_Battery_Module import Generic_Battery_Module  

# ----------------------------------------------------------------------------------------------------------------------
#  Lithium_Air
# ----------------------------------------------------------------------------------------------------------------------   
class Lithium_Air(Generic_Battery_Module):
    """
    Class for modeling lithium-air battery characteristics and performance
    
    Attributes
    ----------
    cell.specific_energy : float
        Energy capacity per unit mass [J/kg] (default: 2000 Wh/kg)
        
    cell.specific_power : float
        Maximum power output per unit mass [W/kg] (default: 0.66 kW/kg)
        
    cell.mass_gain_factor : float
        Rate of mass increase from oxygen absorption [kg/Wh] (default: 1.92E-4)

    Notes
    -----
    Lithium-air batteries use ambient oxygen as the cathode material, which is
    absorbed during discharge, leading to mass gain. They offer very high specific
    energy but have power density limitations.

    See Also
    --------
    RCAIDE.Library.Components.Energy.Sources.Battery_Modules.Generic_Battery_Module
        Base battery module class
    """
    
    def __defaults__(self):
        """
        Sets default values for lithium-air battery attributes
        """      
        self.cell.specific_energy  = 2000.     *Units.Wh/Units.kg    # convert to Joules/kg
        self.cell.specific_power   = 0.66      *Units.kW/Units.kg    # convert to W/kg
        self.cell.mass_gain_factor = (1.92E-4) /Units.Wh


# RCAIDE/Library/Missions/Common/Update/Aerodynamics.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports 
import numpy as np 
    
# ----------------------------------------------------------------------------------------------------------------------
#  Update Freestream
# ----------------------------------------------------------------------------------------------------------------------
def freestream(segment):
    """ Updates the freestream conditions 
        
        Assumptions:
        N/A
        
        Inputs:
            segment.state.conditions.:
                 frames.inertial.velocity_vector   [m/s]
                 freestream.density                [kg/m^3]
                 freestream.speed_of_sound         [m/s]
                 freestream.dynamic_viscosity      [Pa-s]
        Outputs:
            segment.conditions.freestream
                 velocity                          [m/s]
                 mach_number                       [-]
                 reynolds_number                   [-]
                 dynamic_pressure                  [Pa] 
      
        Properties Used:
        N/A
                    
    """   
    
    # unpack
    conditions = segment.state.conditions
    V          = conditions.frames.inertial.velocity_vector
    rho        = conditions.freestream.density
    a          = conditions.freestream.speed_of_sound
    mu         = conditions.freestream.dynamic_viscosity

    # velocity magnitude
    Vmag2 = np.sum( V**2, axis=1)[:,None]  
    Vmag  = np.sqrt(Vmag2)

    # dynamic pressure
    q = 0.5 * rho * Vmag2 # Pa

    # Mach number
    M = Vmag / a

    # Reynolds number
    Re = rho * Vmag / mu  # per m

    # pack
    conditions.freestream.velocity         = Vmag
    conditions.freestream.mach_number      = M
    conditions.freestream.reynolds_number  = Re
    conditions.freestream.dynamic_pressure = q

    return

